"""
Classic A-star algorithm for path finding.
"""

import math

from arcade import (
    BasicSprite,
    Sprite,
    SpriteSequence,
    check_for_collision_with_list,
    get_sprites_at_point,
)
from arcade.math import get_distance, lerp_2d
from arcade.types import Point2

__all__ = ["AStarBarrierList", "astar_calculate_path", "has_line_of_sight"]


def _spot_is_blocked(
    position: Point2, moving_sprite: Sprite, blocking_sprites: SpriteSequence[BasicSprite]
) -> bool:
    """
    Return if position is blocked

    Args:
        position (Point2): position to put moving_sprite at
        moving_sprite (Sprite): Sprite to use
        blocking_sprites (SpriteList): List of Sprites to check against

    Returns:
        bool: If the Sprite would hit anything in blocking_sprites at the position
    """
    original_pos = moving_sprite.position
    moving_sprite.position = position
    hit_list = check_for_collision_with_list(moving_sprite, blocking_sprites)
    moving_sprite.position = original_pos
    return len(hit_list) > 0


def _heuristic(start: Point2, goal: Point2) -> float:
    """
    Returns a heuristic value for the passed points.

    Args:
        start (Point2): The 1st point to compare
        goal (Point2): The 2nd point to compare

    Returns:
        float: The heuristic of the 2 points
    """
    # Use Chebyshev distance heuristic if we can move one square either
    # adjacent or diagonal
    d = 1
    d2 = 1
    dx = abs(start[0] - goal[0])
    dy = abs(start[1] - goal[1])
    return d * (dx + dy) + (d2 - 2 * d) * min(dx, dy)


class _AStarGraph:
    """
    A grid which tracks 2 barriers and a moving sprite.

    Args:
        barriers:
            Barriers to use in the AStarSearch Algorithm.
            These are turned into a set.
        left (int):
            Far left side x value
        right (int):
            Far right side x value
        bottom (int):
            Far bottom side y value
        top (int):
            Far top side y value
        diagonal_movement (bool):
            Whether or not to use diagonals in the AStarSearch Algorithm
    """

    def __init__(
        self,
        barriers: list | tuple | set,
        left: int,
        right: int,
        bottom: int,
        top: int,
        diagonal_movement: bool,
    ):
        self.barriers = barriers if barriers is set else set(barriers)
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom

        if diagonal_movement:
            self.movement_directions = (
                (1, 0),
                (-1, 0),
                (0, 1),
                (0, -1),
                (1, 1),
                (-1, 1),
                (1, -1),
                (-1, -1),
            )
        else:
            self.movement_directions = (1, 0), (-1, 0), (0, 1), (0, -1)  # type: ignore

    def get_vertex_neighbours(self, pos: Point2) -> list[tuple[float, float]]:
        """
        Return neighbors for this point according to ``self.movement_directions``

        These are not guaranteed to be reachable or valid points.

        Args:
            pos: Which position to search around
        Returns:
            Vertices around the point
        """
        n = []
        # Moves allow link a chess king
        for dx, dy in self.movement_directions:
            x2 = pos[0] + dx
            y2 = pos[1] + dy
            if x2 < self.left or x2 > self.right or y2 < self.bottom or y2 > self.top:
                continue
            n.append((x2, y2))
        return n

    def move_cost(self, a: Point2, b: Point2) -> float:
        """
        Returns a float of the cost to move

        Moving diagonally costs more than to the side.
        A barrier's cost is float("inf) so that that
        the Algorithm will never go on it

        Args:
            a: The 1st point to compare
            b: The 2nd point to compare
        Returns:
            The move cost of moving between of the 2 points
        """
        if b in self.barriers:
            return float("inf")  # Infinitely high cost to enter barrier squares

        elif a[0] == b[0] or a[1] == b[1]:
            return 1  # Normal movement cost
        else:
            return 1.42


def _AStarSearch(start: Point2, end: Point2, graph: _AStarGraph) -> list[Point2] | None:
    """
    Returns a path from start to end using the AStarSearch Algorithm

    Graph is used to check for barriers.

    Args:
        start: point to start at
        end: point to end at
        graph: Graph to use
    Returns:
        The path from start to end. Returns ``None`` if is path is not found
    """
    G: dict[Point2, float] = dict()  # Actual movement cost to each position from the start position
    F: dict[Point2, float] = (
        dict()
    )  # Estimated movement cost of start to end going via this position

    # Initialize starting values
    G[start] = 0
    F[start] = _heuristic(start, end)

    closed_vertices = set()
    open_vertices = {start}  # type: ignore
    came_from = {}  # type: ignore

    count = 0
    while len(open_vertices) > 0:
        count += 1
        if count > 500:
            break
        # Get the vertex in the open list with the lowest F score
        current = None
        current_fscore = math.inf
        for pos in sorted(open_vertices):
            if current is None or F[pos] < current_fscore:
                current_fscore = F[pos]
                current = pos

        # Check if we have reached the goal
        if current == end and current is not None:
            # Retrace our route backward
            path = [current]
            while current in came_from:
                current = came_from[current]
                path.append(current)  # type: ignore
            path.reverse()
            if F[end] >= 10000:
                return None
            else:
                return path
            # return path, F[end]  # Done!

        # Mark the current vertex as closed
        open_vertices.remove(current)  # type: ignore
        closed_vertices.add(current)  # type: ignore

        # Update scores for vertices near the current position
        for neighbour in sorted(graph.get_vertex_neighbours(current)):  # type: ignore
            if neighbour in closed_vertices:
                continue  # We have already processed this node exhaustively
            candidate_g = G[current] + graph.move_cost(current, neighbour)  # type: ignore

            if neighbour not in open_vertices:
                open_vertices.add(neighbour)  # Discovered a new vertex
            elif candidate_g >= G[neighbour]:
                continue  # This G score is worse than previously found

            # Adopt this G score
            came_from[neighbour] = current
            G[neighbour] = candidate_g
            h = _heuristic(neighbour, end)
            F[neighbour] = G[neighbour] + h

    # Out-of-bounds
    return None


def _collapse(pos: Point2, grid_size: float) -> tuple[int, int]:
    """Makes Point pos smaller by grid_size"""
    return int(pos[0] // grid_size), int(pos[1] // grid_size)


def _expand(pos: Point2, grid_size: float) -> tuple[int, int]:
    """Makes Point pos larger by grid_size"""
    return int(pos[0] * grid_size), int(pos[1] * grid_size)


class AStarBarrierList:
    """
    Class that manages a list of barriers that can be encountered during
    A* path finding.

    Args:
        moving_sprite:
            Sprite that will be moving
        blocking_sprites:
            Sprites that can block movement
        grid_size (int):
            Size of the grid, in pixels
        left (int):
            Left border of playing field
        right (int):
            Right border of playing field
        bottom (int):
            Bottom of playing field
        top (int):
            Top of playing field

    Attributes:
        grid_size:
            Grid size
        bottom:
            Bottom of playing field
        top:
            Top of playing field
        left:
            Left border of playing field
        right:
            Right border of playing field
        moving_sprite:
            Sprite that will be moving
        blocking_sprites:
            Sprites that can block movement
        barrier_list:
            SpriteList of barriers to use in _AStarSearch,
            ``None`` if not recalculated
    """

    def __init__(
        self,
        moving_sprite: Sprite,
        blocking_sprites: SpriteSequence[BasicSprite],
        grid_size: int,
        left: int,
        right: int,
        bottom: int,
        top: int,
    ):
        self.grid_size = grid_size
        self.bottom = int(bottom // grid_size)
        self.top = int(top // grid_size)
        self.left = int(left // grid_size)
        self.right = int(right // grid_size)
        self.moving_sprite = moving_sprite
        self.blocking_sprites = blocking_sprites
        self.barrier_list = None

        self.recalculate()

    def recalculate(self):
        """Recalculate blocking sprites."""
        # --- Iterate through the blocking sprites and find where we are blocked

        # Save original location
        original_pos = self.moving_sprite.position
        # Create a set of barriers
        self.barrier_list = set()
        # Loop through the grid
        for cx in range(self.left, self.right + 1):
            for cy in range(self.bottom, self.top + 1):
                # Grid location
                cpos = cx, cy
                # Pixel location
                pos = _expand(cpos, self.grid_size)

                # See if we'll have a collision if our sprite is at this location
                self.moving_sprite.position = pos
                if (
                    len(check_for_collision_with_list(self.moving_sprite, self.blocking_sprites))
                    > 0
                ):
                    self.barrier_list.add(cpos)

        # Restore original location
        self.moving_sprite.position = original_pos
        self.barrier_list = sorted(self.barrier_list)


def astar_calculate_path(
    start_point: Point2,
    end_point: Point2,
    astar_barrier_list: AStarBarrierList,
    diagonal_movement: bool = True,
) -> list[Point2] | None:
    """
    Calculates the path using AStarSearch Algorithm and returns the path

    Args:
        start_point:
            Where it starts
        end_point:
            Where it ends
        astar_barrier_list:
            AStarBarrierList with the boundaries to use in the AStarSearch Algorithm
        diagonal_movement:
            Whether of not to use diagonals in the AStarSearch Algorithm

    Returns:
        List of points (the path), or ``None`` if no path is found
    """

    grid_size = astar_barrier_list.grid_size
    mod_start = _collapse(start_point, grid_size)
    mod_end = _collapse(end_point, grid_size)

    left = astar_barrier_list.left
    right = astar_barrier_list.right

    bottom = astar_barrier_list.bottom
    top = astar_barrier_list.top

    barrier_list = astar_barrier_list.barrier_list

    graph = _AStarGraph(barrier_list, left, right, bottom, top, diagonal_movement)  # type: ignore
    result = _AStarSearch(mod_start, mod_end, graph)

    if result is None:
        return None

    # Currently 'result' is in grid locations. We need to convert them to pixel
    # locations.
    revised_result: list[Point2] = [_expand(p, grid_size) for p in result]
    return revised_result


def has_line_of_sight(
    observer: Point2,
    target: Point2,
    walls: SpriteSequence[BasicSprite],
    max_distance: float = float("inf"),
    check_resolution: int = 2,
) -> bool:
    """
    Determine if we have line of sight between two points.

    .. warning:: Try to make sure spatial hashing is enabled on ``walls``!

        If spatial hashing is not enabled, this function may run
        very slowly!

    Args:
        observer:
            Start position
        target:
            End position position
        walls:
            List of all blocking sprites
        max_distance:
            Max distance point 1 can see
        check_resolution:
            Check every x pixels for a sprite.
            Trade-off between accuracy and speed.

    Returns:
        Whether or not observer to target is blocked by any wall in walls
    """
    if max_distance <= 0:
        raise ValueError("max_distance must be greater than zero")
    if check_resolution <= 0:
        raise ValueError("check_resolution must be greater than zero")

    distance = get_distance(observer[0], observer[1], target[0], target[1])
    if distance == 0:
        return True

    steps = int(distance // check_resolution)
    for step in range(steps + 1):
        step_distance = step * check_resolution
        u = step_distance / distance
        midpoint = lerp_2d(observer, target, u)
        if step_distance > max_distance:
            return False
        sprite_list = get_sprites_at_point(midpoint, walls)
        if len(sprite_list) > 0:
            return False
    return True


# NOTE: Rewrite this
# def dda_step(start: Point2, end: Point2):
#     """
#     Bresenham's line algorithm

#     """
#     x1, y1 = start
#     x2, y2 = end

#     dx = x2 - x1
#     dy = y2 - y1

#     steps = max(abs(dx), abs(dy))

#     x_inc = dx / steps
#     y_inc = dy / steps

#     x = x1
#     y = y1

#     points = []
#     for _ in range(steps):
#         points.append((int(x), int(y)))
#         x += x_inc
#         y += y_inc

#     return points



"""
Show how to use lights.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.light_demo
"""
import arcade
from arcade.future.light import Light, LightLayer

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Lighting Demo"
MOVEMENT_SPEED = 5

VIEWPORT_MARGIN = 200
HORIZONTAL_BOUNDARY = WINDOW_WIDTH / 2.0 - VIEWPORT_MARGIN
VERTICAL_BOUNDARY = WINDOW_HEIGHT / 2.0 - VIEWPORT_MARGIN
# If the player moves further than this boundary away from
# the camera we use a constraint to move the camera
CAMERA_BOUNDARY = arcade.LRBT(
    -HORIZONTAL_BOUNDARY,
    HORIZONTAL_BOUNDARY,
    -VERTICAL_BOUNDARY,
    VERTICAL_BOUNDARY,
)

# This is the color used for 'ambient light'. If you don't want any
# ambient light, set it to black.
AMBIENT_COLOR = (10, 10, 10, 255)


class GameView(arcade.View):

    def __init__(self):
        """ Set up the class. """
        super().__init__()

        # Sprite lists
        self.background_sprite_list = None
        self.player_list = None
        self.wall_list = None
        self.player_sprite = None

        # Physics engine
        self.physics_engine = None

        # Camera
        self.camera: arcade.Camera2D = None

        # --- Light related ---
        # List of all the lights
        self.light_layer = None
        # Individual light we move with player, and turn on/off
        self.player_light = None

    def setup(self):
        """ Create everything """

        # Create camera
        self.camera = arcade.Camera2D()

        # Create sprite lists
        self.background_sprite_list = arcade.SpriteList()
        self.player_list = arcade.SpriteList()
        self.wall_list = arcade.SpriteList()

        # Create player sprite
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=0.4)
        self.player_sprite.center_x = 64
        self.player_sprite.center_y = 270
        self.player_list.append(self.player_sprite)

        # --- Light related ---
        # Lights must shine on something. If there is no background sprite or color,
        # you will just see black. Therefore, we use a loop to create a whole
        # bunch of brick tiles to go in the background.
        for x in range(-128, 2000, 128):
            for y in range(-128, 1000, 128):
                sprite = arcade.Sprite(":resources:images/tiles/brickTextureWhite.png")
                sprite.position = x, y
                self.background_sprite_list.append(sprite)

        # Create a light layer, used to render things to, then post-process and
        # add lights. This must match the screen size.
        self.light_layer = LightLayer(WINDOW_WIDTH, WINDOW_HEIGHT)
        # We can also set the background color that will be lit by lights,
        # but in this instance we just want a black background
        self.light_layer.set_background_color(arcade.color.BLACK)

        # Here we create a bunch of lights.

        # Create a small white light
        x = 100
        y = 200
        radius = 100
        mode = 'soft'
        color = arcade.csscolor.WHITE
        light = Light(x, y, radius, color, mode)
        self.light_layer.add(light)

        # Create an overlapping, large white light
        x = 300
        y = 150
        radius = 200
        color = arcade.csscolor.WHITE
        mode = 'soft'
        light = Light(x, y, radius, color, mode)
        self.light_layer.add(light)

        # Create three, non-overlapping RGB lights
        x = 50
        y = 450
        radius = 100
        mode = 'soft'
        color = arcade.csscolor.RED
        light = Light(x, y, radius, color, mode)
        self.light_layer.add(light)

        x = 250
        y = 450
        radius = 100
        mode = 'soft'
        color = arcade.csscolor.GREEN
        light = Light(x, y, radius, color, mode)
        self.light_layer.add(light)

        x = 450
        y = 450
        radius = 100
        mode = 'soft'
        color = arcade.csscolor.BLUE
        light = Light(x, y, radius, color, mode)
        self.light_layer.add(light)

        # Create three, overlapping RGB lights
        x = 650
        y = 450
        radius = 100
        mode = 'soft'
        color = arcade.csscolor.RED
        light = Light(x, y, radius, color, mode)
        self.light_layer.add(light)

        x = 750
        y = 450
        radius = 100
        mode = 'soft'
        color = arcade.csscolor.GREEN
        light = Light(x, y, radius, color, mode)
        self.light_layer.add(light)

        x = 850
        y = 450
        radius = 100
        mode = 'soft'
        color = arcade.csscolor.BLUE
        light = Light(x, y, radius, color, mode)
        self.light_layer.add(light)

        # Create three, overlapping RGB lights
        # But 'hard' lights that don't fade out.
        x = 650
        y = 150
        radius = 100
        mode = 'hard'
        color = arcade.csscolor.RED
        light = Light(x, y, radius, color, mode)
        self.light_layer.add(light)

        x = 750
        y = 150
        radius = 100
        mode = 'hard'
        color = arcade.csscolor.GREEN
        light = Light(x, y, radius, color, mode)
        self.light_layer.add(light)

        x = 850
        y = 150
        radius = 100
        mode = 'hard'
        color = arcade.csscolor.BLUE
        light = Light(x, y, radius, color, mode)
        self.light_layer.add(light)

        # Create a light to follow the player around.
        # We'll position it later, when the player moves.
        # We'll only add it to the light layer when the player turns the light
        # on. We start with the light off.
        radius = 150
        mode = 'soft'
        color = arcade.csscolor.WHITE
        self.player_light = Light(0, 0, radius, color, mode)

        # Create the physics engine
        self.physics_engine = arcade.PhysicsEngineSimple(self.player_sprite, self.wall_list)

    def on_draw(self):
        """ Draw everything. """
        self.clear()

        # --- Light related ---
        # Everything that should be affected by lights gets rendered inside this
        # 'with' statement. Nothing is rendered to the screen yet, just the light
        # layer.
        with self.light_layer:
            self.background_sprite_list.draw()
            self.player_list.draw()

        # Draw the light layer to the screen.
        # This fills the entire screen with the lit version
        # of what we drew into the light layer above.
        self.light_layer.draw(ambient_color=AMBIENT_COLOR)

        # Now draw anything that should NOT be affected by lighting.
        left, bottom = self.camera.bottom_left
        arcade.draw_text("Press SPACE to turn character light on/off.",
                         10 + int(left), 10 + int(bottom),
                         arcade.color.WHITE, 20)

    def on_resize(self, width, height):
        """ User resizes the screen. """
        super().on_resize(width, height)
        self.camera.match_window()

        # --- Light related ---
        # We need to resize the light layer to
        self.light_layer.resize(width, height)

        # Scroll the screen so the user is visible
        self.scroll_screen()

    def on_key_press(self, key, _):
        """Called whenever a key is pressed. """

        if key == arcade.key.UP:
            self.player_sprite.change_y = MOVEMENT_SPEED
        elif key == arcade.key.DOWN:
            self.player_sprite.change_y = -MOVEMENT_SPEED
        elif key == arcade.key.LEFT:
            self.player_sprite.change_x = -MOVEMENT_SPEED
        elif key == arcade.key.RIGHT:
            self.player_sprite.change_x = MOVEMENT_SPEED
        elif key == arcade.key.SPACE:
            # --- Light related ---
            # We can add/remove lights from the light layer. If they aren't
            # in the light layer, the light is off.
            if self.player_light in self.light_layer:
                self.light_layer.remove(self.player_light)
            else:
                self.light_layer.add(self.player_light)

    def on_key_release(self, key, _):
        """Called when the user releases a key. """

        if key == arcade.key.UP or key == arcade.key.DOWN:
            self.player_sprite.change_y = 0
        elif key == arcade.key.LEFT or key == arcade.key.RIGHT:
            self.player_sprite.change_x = 0

    def scroll_screen(self):
        """ Manage Scrolling """

        # --- Manage Scrolling ---
        self.camera.position = arcade.camera.grips.constrain_boundary_xy(
            self.camera.view_data, CAMERA_BOUNDARY, self.player_sprite.position
        )

        self.camera.use()

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.physics_engine.update()

        # --- Light related ---
        # We can easily move the light by setting the position,
        # or by center_x, center_y.
        self.player_light.position = self.player_sprite.position

        # Scroll the screen so we can see the player
        self.scroll_screen()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
GPU version of bouncing coins.

The downside of manipulating spritelists
with shaders is that the data in the sprites
are stale, but it can still be a useful tool.

GPU collision checking can potentially be used if needed.

The window can be resized.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.spritelist_interaction_bouncing_coins
"""

from array import array
from random import randint, uniform

import arcade
from arcade.gl.types import BufferDescription
from arcade import hitbox

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720

NUM_COINS = 10_000


class GPUBouncingCoins(arcade.Window):

    def __init__(self):
        super().__init__(WINDOW_WIDTH, WINDOW_HEIGHT, resizable=True)

        # Generate lots of coins in random positions
        self.coins = arcade.SpriteList(use_spatial_hash=None)
        texture = arcade.load_texture(
            ":resources:images/items/coinGold.png",
            hit_box_algorithm=hitbox.algo_bounding_box,
        )
        for _ in range(NUM_COINS):
            self.coins.append(
                arcade.Sprite(
                    texture,
                    scale=0.25,
                    center_x=randint(0, WINDOW_WIDTH),
                    center_y=randint(0, WINDOW_HEIGHT),
                )
            )
        # Ensure internal buffer data are up to date
        self.coins.write_sprite_buffers_to_gpu()

        # Transform shader writing each out value to separate buffers.
        # We're simply moving the sprites based on velocity and
        # reversing direction if outside the screen
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            uniform float delta_time;
            uniform vec2 size;

            in vec3 in_pos;
            in vec2 in_vel;

            out vec3 out_pos;
            out vec2 out_vel;

            void main() {
                vec2 pos = in_pos.xy + in_vel * 100.0 * delta_time;
                vec2 vel = in_vel;
                if (pos.x > size.x) {
                    pos.x = size.x;
                    vel.x *= -1.0;
                }
                else if (pos.x < 0.0) {
                    pos.x = 0.0;
                    vel.x *= -1.0;
                }
                if (pos.y > size.y) {
                    pos.y = size.y;
                    vel.y *= -1.0;
                }
                else if (pos.t < 0.0) {
                    pos.y = 0.0;
                    vel.y *= -1.0;
                }
                out_pos = vec3(pos, in_pos.z);
                out_vel = vel;
            }
            """,
            varyings_capture_mode="separate",
        )
        # We need two position buffers and two velocity buffers
        # because we can't safely write to buffers we are reading from.
        # We create two velocity buffers because spritelist don't have velocity buffers
        # We only create one position buffer because the second one is in the spritelist

        # Buffer with some quick random velocities for our coins
        self.buffer_velocity_1 = self.ctx.buffer(
            data=array("f", [uniform(-1, 1) for _ in range(NUM_COINS * 2)])
        )
        # Second velocity buffer
        self.buffer_velocity_2 = self.ctx.buffer(reserve=self.buffer_velocity_1.size)
        # Create a buffer with the same size as the position buffer in  the spritelist.
        # It's important that these match because we're copying that buffer into this one.
        self.buffer_pos_copy = self.ctx.buffer(reserve=self.coins.buffer_positions.size)

        # Geometry input: Copied positions and first velocity buffer
        self.geometry_1 = self.ctx.geometry(
            [
                BufferDescription(self.buffer_pos_copy, "3f", ["in_pos"]),
                BufferDescription(self.buffer_velocity_1, "2f", ["in_vel"]),
            ]
        )
        # Geometry input: Copied positions and second velocity buffer
        self.geometry_2 = self.ctx.geometry(
            [
                BufferDescription(self.buffer_pos_copy, "3f", ["in_pos"]),
                BufferDescription(self.buffer_velocity_2, "2f", ["in_vel"]),
            ]
        )

    def on_draw(self):
        self.clear()

        # Copy the position buffer. This happens on the gpu side.
        self.buffer_pos_copy.copy_from_buffer(self.coins.buffer_positions)

        # Run the transform writing new positions and velocities
        self.geometry_1.transform(
            self.program,
            [
                self.coins.buffer_positions,
                self.buffer_velocity_2,
            ],
        )

        self.coins.draw()

        # Swap things around for next frame
        (self.buffer_velocity_1, self.buffer_velocity_2) = self.buffer_velocity_2, self.buffer_velocity_1  # noqa
        self.geometry_1, self.geometry_2 = self.geometry_2, self.geometry_1

    def on_update(self, delta_time: float):
        self.program["delta_time"] = delta_time
        self.program["size"] = self.get_size()


GPUBouncingCoins().run()



from __future__ import annotations

import weakref
from ctypes import byref, c_uint32
from typing import TYPE_CHECKING

from pyglet import gl

from .types import PyGLuint, compare_funcs

if TYPE_CHECKING:
    from arcade.gl import Context, Texture2D


class Sampler:
    """
    OpenGL sampler object.

    When bound to a texture unit it overrides all the
    sampling parameters of the texture channel.
    """

    def __init__(
        self,
        ctx: "Context",
        texture: Texture2D,
        *,
        filter: tuple[PyGLuint, PyGLuint] | None = None,
        wrap_x: PyGLuint | None = None,
        wrap_y: PyGLuint | None = None,
    ):
        self._ctx = ctx
        self._glo = -1

        value = c_uint32()
        gl.glGenSamplers(1, byref(value))
        self._glo = value.value

        self.texture = texture

        # Default filters for float and integer textures
        # Integer textures should have NEAREST interpolation
        # by default 3.3 core doesn't really support it consistently.
        if "f" in self.texture._dtype:
            self._filter = gl.GL_LINEAR, gl.GL_LINEAR
        else:
            self._filter = gl.GL_NEAREST, gl.GL_NEAREST

        self._wrap_x = gl.GL_REPEAT
        self._wrap_y = gl.GL_REPEAT
        self._anisotropy = 1.0
        self._compare_func: str | None = None

        # Only set texture parameters on non-multisample textures
        if self.texture._samples == 0:
            self.filter = filter or self._filter
            self.wrap_x = wrap_x or self._wrap_x
            self.wrap_y = wrap_y or self._wrap_y

        if self._ctx.gc_mode == "auto":
            weakref.finalize(self, Sampler.delete_glo, self._glo)

    @property
    def glo(self) -> PyGLuint:
        """The OpenGL sampler id"""
        return self._glo

    def use(self, unit: int):
        """
        Bind the sampler to a texture unit
        """
        gl.glBindSampler(unit, self._glo)

    def clear(self, unit: int):
        """
        Unbind the sampler from a texture unit
        """
        gl.glBindSampler(unit, 0)

    @property
    def filter(self) -> tuple[int, int]:
        """
        Get or set the ``(min, mag)`` filter for this texture.

        These are rules for how a texture interpolates.
        The filter is specified for minification and magnification.

        Default value is ``LINEAR, LINEAR``.
        Can be set to ``NEAREST, NEAREST`` for pixelated graphics.

        When mipmapping is used the min filter needs to be one of the
        ``MIPMAP`` variants.

        Accepted values::

            # Enums can be accessed on the context or arcade.gl
            NEAREST                # Nearest pixel
            LINEAR                 # Linear interpolate
            NEAREST_MIPMAP_NEAREST # Minification filter for mipmaps
            LINEAR_MIPMAP_NEAREST  # Minification filter for mipmaps
            NEAREST_MIPMAP_LINEAR  # Minification filter for mipmaps
            LINEAR_MIPMAP_LINEAR   # Minification filter for mipmaps

        Also see

        * https://www.khronos.org/opengl/wiki/Texture#Mip_maps
        * https://www.khronos.org/opengl/wiki/Sampler_Object#Filtering
        """
        return self._filter

    @filter.setter
    def filter(self, value: tuple[int, int]):
        if not isinstance(value, tuple) or not len(value) == 2:
            raise ValueError("Texture filter must be a 2 component tuple (min, mag)")

        self._filter = value
        gl.glSamplerParameteri(self._glo, gl.GL_TEXTURE_MIN_FILTER, self._filter[0])
        gl.glSamplerParameteri(self._glo, gl.GL_TEXTURE_MAG_FILTER, self._filter[1])

    @property
    def wrap_x(self) -> int:
        """
        Get or set the horizontal wrapping of the texture.

        This decides how textures are read when texture coordinates are outside
        the ``[0.0, 1.0]`` area. Default value is ``REPEAT``.

        Valid options are::

            # Note: Enums can also be accessed in arcade.gl
            # Repeat pixels on the y axis
            texture.wrap_x = ctx.REPEAT
            # Repeat pixels on the y axis mirrored
            texture.wrap_x = ctx.MIRRORED_REPEAT
            # Repeat the edge pixels when reading outside the texture
            texture.wrap_x = ctx.CLAMP_TO_EDGE
            # Use the border color (black by default) when reading outside the texture
            texture.wrap_x = ctx.CLAMP_TO_BORDER
        """
        return self._wrap_x

    @wrap_x.setter
    def wrap_x(self, value: int):
        self._wrap_x = value
        gl.glSamplerParameteri(self._glo, gl.GL_TEXTURE_WRAP_S, value)

    @property
    def wrap_y(self) -> int:
        """
        Get or set the horizontal wrapping of the texture.

        This decides how textures are read when texture coordinates are outside the
        ``[0.0, 1.0]`` area. Default value is ``REPEAT``.

        Valid options are::

            # Note: Enums can also be accessed in arcade.gl
            # Repeat pixels on the x axis
            texture.wrap_x = ctx.REPEAT
            # Repeat pixels on the x axis mirrored
            texture.wrap_x = ctx.MIRRORED_REPEAT
            # Repeat the edge pixels when reading outside the texture
            texture.wrap_x = ctx.CLAMP_TO_EDGE
            # Use the border color (black by default) when reading outside the texture
            texture.wrap_x = ctx.CLAMP_TO_BORDER
        """
        return self._wrap_y

    @wrap_y.setter
    def wrap_y(self, value: int):
        self._wrap_y = value
        gl.glSamplerParameteri(self._glo, gl.GL_TEXTURE_WRAP_T, value)

    @property
    def anisotropy(self) -> float:
        """Get or set the anisotropy for this texture."""
        return self._anisotropy

    @anisotropy.setter
    def anisotropy(self, value):
        self._anisotropy = max(1.0, min(value, self._ctx.info.MAX_TEXTURE_MAX_ANISOTROPY))
        gl.glSamplerParameterf(self._glo, gl.GL_TEXTURE_MAX_ANISOTROPY, self._anisotropy)

    @property
    def compare_func(self) -> str | None:
        """
        Get or set the compare function for a depth texture::

            texture.compare_func = None  # Disable depth comparison completely
            texture.compare_func = '<='  # GL_LEQUAL
            texture.compare_func = '<'   # GL_LESS
            texture.compare_func = '>='  # GL_GEQUAL
            texture.compare_func = '>'   # GL_GREATER
            texture.compare_func = '=='  # GL_EQUAL
            texture.compare_func = '!='  # GL_NOTEQUAL
            texture.compare_func = '0'   # GL_NEVER
            texture.compare_func = '1'   # GL_ALWAYS
        """
        return self._compare_func

    @compare_func.setter
    def compare_func(self, value: str | None):
        if not self.texture._depth:
            raise ValueError("Depth comparison function can only be set on depth textures")

        if not isinstance(value, str) and value is not None:
            raise ValueError(f"value must be as string: {self._compare_funcs.keys()}")

        func = compare_funcs.get(value, None)
        if func is None:
            raise ValueError(f"value must be as string: {compare_funcs.keys()}")

        self._compare_func = value
        if value is None:
            gl.glSamplerParameteri(self._glo, gl.GL_TEXTURE_COMPARE_MODE, gl.GL_NONE)
        else:
            gl.glSamplerParameteri(
                self._glo, gl.GL_TEXTURE_COMPARE_MODE, gl.GL_COMPARE_REF_TO_TEXTURE
            )
            gl.glSamplerParameteri(self._glo, gl.GL_TEXTURE_COMPARE_FUNC, func)

    @staticmethod
    def delete_glo(glo: int) -> None:
        """
        Delete the OpenGL object
        """
        gl.glDeleteSamplers(1, glo)



import array
import math

from arcade import gl
from arcade.types import Color, Point2, Point2List, RGBOrA255
from arcade.window_commands import get_window


def get_points_for_thick_line(
    start_x: float, start_y: float, end_x: float, end_y: float, line_width: float
) -> tuple[Point2, Point2, Point2, Point2]:
    """
    Function used internally for Arcade. OpenGL draws triangles only, so a thick
    line must be two triangles that make up a rectangle. This calculates and returns
    those points.
    """
    vector_x = start_x - end_x
    vector_y = start_y - end_y
    perpendicular_x = vector_y
    perpendicular_y = -vector_x
    length = math.sqrt(vector_x * vector_x + vector_y * vector_y)
    if length == 0:
        normal_x = 1.0
        normal_y = 1.0
    else:
        normal_x = perpendicular_x / length
        normal_y = perpendicular_y / length

    half_width = line_width / 2
    shift_x = normal_x * half_width
    shift_y = normal_y * half_width

    r1_x = start_x + shift_x
    r1_y = start_y + shift_y
    r2_x = start_x - shift_x
    r2_y = start_y - shift_y
    r3_x = end_x + shift_x
    r3_y = end_y + shift_y
    r4_x = end_x - shift_x
    r4_y = end_y - shift_y

    return (r1_x, r1_y), (r2_x, r2_y), (r4_x, r4_y), (r3_x, r3_y)


def _generic_draw_line_strip(
    point_list: Point2List, color: RGBOrA255, mode: int = gl.LINE_STRIP
) -> None:
    """
    Draw a line strip. A line strip is a set of continuously connected
    line segments.

    Args:
        point_list:
            List of points making up the line. Each point is in a list.
            So it is a list of lists.
        color:
            A color, specified as an RGBA tuple or a
            :py:class:`.Color` instance.
    """
    # Fail if we don't have a window, context, or right GL abstractions
    window = get_window()
    ctx = window.ctx
    geometry = ctx.generic_draw_line_strip_geometry  # type: ignore
    vertex_buffer = ctx.generic_draw_line_strip_vbo  # type: ignore
    color_buffer = ctx.generic_draw_line_strip_color  # type: ignore
    program = ctx.line_vertex_shader

    # Validate and alpha-pad color, then expand to multi-vertex form since
    # this shader normalizes internally as if made to draw multicolor lines.
    rgba = Color.from_iterable(color)
    num_vertices = len(point_list)  # Fail if it isn't a sized / sequence object

    # Translate Python objects into types Arcade's Buffer objects accept
    color_array = array.array("B", rgba * num_vertices)
    vertex_array = array.array("f", tuple(item for sublist in point_list for item in sublist))
    geometry.num_vertices = num_vertices

    # Double buffer sizes until they can hold all our data
    goal_vertex_buffer_size = len(vertex_array) * 4
    while goal_vertex_buffer_size > vertex_buffer.size:
        vertex_buffer.orphan(color_buffer.size * 2)
        color_buffer.orphan(color_buffer.size * 2)
    else:
        vertex_buffer.orphan()
        color_buffer.orphan()

    ctx.enable(ctx.BLEND)

    # Write data & render
    vertex_buffer.write(vertex_array)
    color_buffer.write(color_array)
    geometry.render(program, mode=mode)

    ctx.disable(ctx.BLEND)



from __future__ import annotations

import contextlib
import copy
from collections.abc import Sequence

# import logging
# import time
from pathlib import Path
from typing import (
    TYPE_CHECKING,
)
from weakref import WeakSet, WeakValueDictionary, finalize, ref

import PIL.Image
from PIL import Image, ImageDraw
from PIL.Image import Resampling
from pyglet.image.atlas import (
    Allocator,
    AllocatorException,
)
from pyglet.math import Mat4

from arcade.camera.static import static_from_raw_orthographic
from arcade.texture.transforms import Transform
from arcade.window_commands import get_window

from .base import TextureAtlasBase
from .ref_counters import (
    ImageDataRefCounter,
    UniqueTextureRefCounter,
)
from .region import AtlasRegion
from .uv_data import UVData

if TYPE_CHECKING:
    from arcade import ArcadeContext, Texture
    from arcade.gl import Texture2D
    from arcade.texture import ImageData

# The amount of pixels we increase the atlas when scanning for a reasonable size.
# It must divide. Must be a power of two number like 64, 256, 512 etx
RESIZE_STEP = 128
# This is the maximum size of the float32 UV texture. 4096 is a safe value for
# OpenGL ES 3.1/2. It's not recommended to go higher than this. This is a 2D
# texture anyway, so more rows can be added.
UV_TEXTURE_WIDTH = 4096

# LOG = logging.getLogger(__name__)
# LOG.handlers = [logging.StreamHandler()]
# LOG.setLevel(logging.INFO)


class DefaultTextureAtlas(TextureAtlasBase):
    """
    A texture atlas with a size in a context.

    A texture atlas is a large texture containing several textures
    so OpenGL can easily batch draw thousands or hundreds of thousands
    of sprites on one draw operation.

    The atlas is auto-GCed meaning that textures and images are removed
    when the user removes all references to them. This is the only way
    to remove textures and images from the atlas.

    This is a fairly simple atlas that stores horizontal strips were
    the height of the strip is the texture/image with the larges height.
    This is done for performance reasons.

    Adding a texture to this atlas generates a texture id.
    This id is used the sprite list vertex data to reference what
    texture each sprite is using. The actual texture coordinates
    are located in a float32 texture this atlas is responsible for
    keeping up to date.

    The atlas deals with image and textures. The image is the actual
    image data. The texture is the Arcade texture object that contains
    the image and other information about such as transforms.
    Several textures can share the same image with different transforms
    applied. The transforms are simply changing the order of the texture
    coordinates to flip, rotate or mirror the image.

    Args:
        size:
            The width and height of the atlas in pixels
        border:
            The number of edge pixels to repeat around images in the atlas.
            This kind of padding is important to avoid edge artifacts.
            Default is 1 pixel.
        textures:
            Optional sequence of textures to add to the atlas on creation
        auto_resize:
            Automatically resize the atlas when full. Default is ``True``.
        ctx:
            The context for this atlas (will use window context if left empty)
        capacity:
            The number of textures the atlas keeps track of.
            This is multiplied by 4096. Meaning capacity=2 is 8192 textures.
            This value can affect the performance of the atlas.
    """

    def __init__(
        self,
        size: tuple[int, int],
        *,
        border: int = 1,
        textures: Sequence[Texture] | None = None,
        auto_resize: bool = True,
        ctx: ArcadeContext | None = None,
        capacity: int = 2,
    ):
        self._ctx = ctx or get_window().ctx
        self._max_size = self._ctx.info.MAX_VIEWPORT_DIMS
        self._size: tuple[int, int] = size
        self._allocator = Allocator(*self._size)
        self._auto_resize = auto_resize
        self._capacity = capacity
        self._border: int = border
        if self._border < 0:
            raise ValueError("Border must be 0 or a positive integer")

        # Decides the number of texture and images the atlas can hold.
        # Must be a multiple of UV_TEXTURE_WIDTH due texture coordinates being
        # stored in a float32 texture.
        if not isinstance(capacity, int) or capacity < 1:
            raise ValueError("Capacity must be a positive integer")

        self._check_size(self._size)

        # The atlas texture
        self._texture = self._ctx.texture(
            size,
            components=4,
            wrap_x=self._ctx.CLAMP_TO_EDGE,
            wrap_y=self._ctx.CLAMP_TO_EDGE,
        )
        # Creating an fbo makes us able to clear the texture or parts
        # of the texture including rendering to a part of the texture.
        # This also means we can resize the atlas in the gpu
        # by rendering the old atlas into the new one.
        self._fbo = self._ctx.framebuffer(color_attachments=[self._texture])

        # Texture coordinate data for images and textures.
        # * The image UVs are used when rebuilding the atlas
        # * The texture UVs are passed into sprite shaders as a source for texture coordinates
        self._image_uvs = UVData(self._ctx, capacity)
        self._texture_uvs = UVData(self._ctx, capacity)

        # A dictionary of all the allocated regions for images/textures in the atlas.
        # The texture regions are clones of the image regions with transforms applied
        # in order to map the same image using different orders or texture coordinates.
        # The key is the cache name for a texture
        self._image_regions: dict[str, AtlasRegion] = dict()
        self._texture_regions: dict[str, AtlasRegion] = dict()

        # Ref counter for images and textures. Per atlas we need to keep
        # track of ho many times an image is used in textures to determine
        # when to remove an image from the atlas. We only ever track images
        # using their sha256 hash to avoid writing the same image multiple times.
        self._image_ref_count = ImageDataRefCounter()
        # Tracks how many instances of a unique texture we have in the atlas.
        # This basically means textures with the same image and vertex order (atlas_name)
        self._unique_texture_ref_count = UniqueTextureRefCounter()

        # A list of all the images this atlas contains.
        # Unique by: Internal hash property
        self._images: WeakValueDictionary[str, ImageData] = WeakValueDictionary()

        # All textures added to the atlas
        self._textures: WeakSet[Texture] = WeakSet()
        # atlas_name: Set of textures with matching atlas name
        self._unique_textures: dict[str, WeakSet[Texture]] = dict()

        self._textures_added = 0
        self._textures_removed = 0
        self._finalizers_created = 0

        for tex in textures or []:
            self.add(tex)

    @property
    def max_width(self) -> int:
        """The maximum width of the atlas in pixels."""
        return self._max_size[0]

    @property
    def max_height(self) -> int:
        """The maximum height of the atlas in pixels."""
        return self._max_size[1]

    @property
    def max_size(self) -> tuple[int, int]:
        """The maximum size of the atlas in pixels (x, y)."""
        return self._max_size

    @property
    def auto_resize(self) -> bool:
        """
        Get or set the auto resize flag for the atlas.
        If enabled the atlas will resize itself when full.
        """
        return self._auto_resize

    @auto_resize.setter
    def auto_resize(self, value: bool):
        self._auto_resize = value

    @property
    def border(self) -> int:
        """The texture border in pixels"""
        return self._border

    @property
    def image_uv_texture(self) -> Texture2D:
        """Texture coordinate texture for images."""
        return self._image_uvs.texture

    @property
    def texture_uv_texture(self) -> Texture2D:
        """Texture coordinate texture for textures."""
        return self._texture_uvs.texture

    @property
    def textures(self) -> list[Texture]:
        """
        All textures instance added to the atlas regardless
        of their internal state. See :py:meth:`unique_textures``
        for textures with unique image data and transformation.
        """
        return list(self._textures)

    @property
    def unique_textures(self) -> list[Texture]:
        """
        All unique textures in the atlas.

        These are textures using an image with the same hash
        and the same vertex order. The full list of all textures
        can be found in :py:meth:`textures`.
        """
        # Grab the first texture from each set
        textures: list[Texture] = []
        # NOTE: keys can drop out of the dict during iteration.
        #       Copy they keys and look up each set to avoid this.
        for key in list(self._unique_textures.keys()):
            tex_set = self._unique_textures.get(key, None)
            # Entry was GCed during iteration
            if tex_set is None:
                continue
            # Corrupt data
            if len(tex_set) == 0:
                raise RuntimeError("Empty set in unique textures")

            textures.append(next(iter(tex_set)))

        return textures

    @property
    def images(self) -> list[ImageData]:
        """
        Return a list of all the images in the atlas.

        A new list is constructed from the internal weak set of images.
        """
        return list(self._images.values())

    def add(self, texture: Texture) -> tuple[int, AtlasRegion]:
        """
        Add a texture to the atlas.

        Add a texture to the atlas.

        Args:
            texture: The texture to add
        Returns:
            texture_id, AtlasRegion tuple
        Raises:
            AllocatorException: If there are no room for the texture
        """
        return self._add(texture)

    def _add(self, texture: Texture, create_finalizer=True) -> tuple[int, AtlasRegion]:
        """
        Internal add method with additional control. We we rebuild the atlas
        we don't want to create finalizers for the texture or they will be
        removed multiple times causing errors.

        Args:
            texture:
                The texture to add
            create_finalizer:
                If a finalizer should be created
        """
        # Quickly handle a texture already having a unique texture in the atlas
        if self.has_unique_texture(texture):
            # Add add references to the duplicate texture
            if not self.has_texture(texture):
                self._add_texture_ref(texture, create_finalizer=create_finalizer)
            slot = self._texture_uvs.get_slot_or_raise(texture.atlas_name)
            region = self.get_texture_region_info(texture.atlas_name)
            return slot, region

        # Add the *image* to the atlas if it's not already there
        if not self.has_image(texture.image_data):
            try:
                # Attempt to allocate space for the image
                x, y, slot, region = self._allocate_image(texture.image_data)
                # Write the pixel data to the atlas texture
                self.write_image(texture.image_data.image, x, y)
            except AllocatorException:
                if not self._auto_resize:
                    raise AllocatorException(
                        f"No more space for image {texture.image_data.hash} "
                        f"size={texture.image.size}. "
                        f"Curr size: {self._size}. "
                        f"Max size: {self._max_size}"
                    )

                # If we have lost regions/images we can try to rebuild the atlas
                removed_image_count = self._image_ref_count.get_total_decref()
                if removed_image_count > 0:
                    self.rebuild()
                    return self._add(texture, create_finalizer=create_finalizer)

                # Double the size of the atlas (capped by max size)
                width = min(self.width * 2, self.max_width)
                height = min(self.height * 2, self.max_height)
                # If the size didn't change we have a problem ..
                if self._size == (width, height):
                    raise

                # Resize the atlas making more room for images
                self.resize((width, height))

                # Recursively try to add the texture again
                return self._add(texture, create_finalizer=create_finalizer)

        # Finally we can register the texture
        self._add_texture_ref(texture, create_finalizer=create_finalizer)
        info = self._allocate_texture(texture)
        return info

    def _add_texture_ref(self, texture: Texture, create_finalizer=True) -> None:
        """
        Add references to the texture and image data.
        including finalizer to remove the texture when it's no longer used.

        Args:
            texture:
                The texture
            create_finalizer:
                If a finalizer should be created
        """
        self._textures.add(texture)
        self._unique_texture_ref_count.inc_ref(texture)
        self._image_ref_count.inc_ref(texture.image_data)

        if create_finalizer:
            atlas_ref = ref(self)

            # NOTE: The finalizer needs to be completely decoupled from the atlas
            #       or it will self-reference and not die unless all the textures in it
            #       are removed. This lead to leaking orphaned atlases when you have
            #       pre-loaded shared textures in multiple atlases.
            def finalizer_callback(atlas_name, hash):
                atlas = atlas_ref()
                if atlas is not None:
                    atlas._remove_texture_by_identifiers(atlas_name, hash)

            finalizer_ref = finalize(
                texture,
                finalizer_callback,
                texture.atlas_name,
                texture.image_data.hash,
            )
            # Don't bother removing texture on program exit
            finalizer_ref.atexit = False
            self._finalizers_created += 1

        self._textures_added += 1

    def remove(self, texture: Texture) -> None:
        """
        Remove a texture from the atlas.

        This is only supported by static atlases. Dynamic atlases with
        garbage collection will remove texture using python's garbage
        collector.

        Args:
            texture: The texture to remove
        """
        raise RuntimeError(
            "The default texture atlas does not support manual removal of textures. "
            "To remove textures from the atlas you must remove all references to the texture "
            "and let the python garbage collector handle the removal."
        )

    def _allocate_texture(self, texture: Texture) -> tuple[int, AtlasRegion]:
        """
        Add or update a unique texture in the atlas.
        This is mainly responsible for updating the texture coordinates.

        Args:
            texture: The texture to add
        """
        # NOTE: This is also called when re-building the atlas meaning we
        #       need to support updating the texture coordinates for existing textures
        slot = self._texture_uvs.get_existing_or_free_slot(texture.atlas_name)

        # Copy the region for the image and apply the texture transform
        image_region = self.get_image_region_info(texture.image_data.hash)
        texture_region = copy.deepcopy(image_region)
        texture_region.texture_coordinates = Transform.transform_texture_coordinates_order(
            texture_region.texture_coordinates, texture._vertex_order
        )
        self._texture_regions[texture.atlas_name] = texture_region  # add or update region

        # Put texture coordinates into uv buffer
        self._texture_uvs.set_slot_data(slot, texture_region.texture_coordinates)
        # Collect unique textures
        self._unique_textures.setdefault(texture.atlas_name, WeakSet()).add(texture)

        return slot, texture_region

    def _allocate_image(self, image_data: ImageData) -> tuple[int, int, int, AtlasRegion]:
        """
        Attempts to allocate space for an image in the atlas or
        update the existing space for the image.

        This doesn't write the texture to the atlas texture itself.
        It only allocates space.

        Returns:
            The x, y texture_id, TextureRegion
        """
        image = image_data.image

        # Allocate space for texture
        try:
            x, y = self._allocator.alloc(
                image.width + self._border * 2,
                image.height + self._border * 2,
            )
        except AllocatorException:
            raise AllocatorException(
                f"No more space for image {image_data.hash} size={image.size}. "
                f"Curr size: {self._size}. "
                f"Max size: {self._max_size}"
            )

        # LOG.debug("Allocated new space for image %s : %s %s", image_data.hash, x, y)

        # Store a texture region for this allocation
        # The xy position must be offset by the border size
        # while the image size must stay as its true size
        region = AtlasRegion(
            self,
            x + self._border,
            y + self._border,
            image.width,
            image.height,
        )
        self._image_regions[image_data.hash] = region

        # Get the existing slot for this texture or grab a new one.
        # Existing slots for textures will only happen when re-building
        # the atlas since we want to keep the same slots to avoid
        # re-building the sprite list
        slot = self._image_uvs.get_existing_or_free_slot(image_data.hash)
        # Put texture coordinates into uv buffer
        self._image_uvs.set_slot_data(slot, region.texture_coordinates)

        self._images[image_data.hash] = image_data
        return x, y, slot, region

    def write_image(self, image: PIL.Image.Image, x: int, y: int) -> None:
        """
        Write a PIL image to the atlas in a specific region.

        Args:
            image:
                The pillow image
            x:
                The x position to write the texture
            y:
                The y position to write the texture
        """
        # Write into atlas at the allocated location + border
        viewport = (
            x,
            y,
            image.width + self._border * 2,
            image.height + self._border * 2,
        )

        # Only do extrusion if we have a border
        if self._border > 0:
            # Make new image with room for borders
            tmp = Image.new(
                "RGBA",
                size=(image.width + self._border * 2, image.height + self._border * 2),
                color=(0, 0, 0, 0),
            )
            # Paste the image into the center of the new image
            tmp.paste(image, (self._border, self._border))

            # Copy 1 pixel strips from each side of the image to the border
            # so we can repeat this pixel data in the border region
            # left, top, right, bottom
            strip_top = image.crop((0, 0, image.width, 1))
            strip_bottom = image.crop((0, image.height - 1, image.width, image.height))
            strip_left = image.crop((0, 0, 1, image.height))
            strip_right = image.crop((image.width - 1, 0, image.width, image.height))

            # Resize the strips to the border size if larger than 1
            if self._border > 1:
                strip_top = strip_top.resize((image.width, self._border), Resampling.NEAREST)
                strip_bottom = strip_bottom.resize((image.width, self._border), Resampling.NEAREST)
                strip_left = strip_left.resize((self._border, image.height), Resampling.NEAREST)
                strip_right = strip_right.resize((self._border, image.height), Resampling.NEAREST)

            tmp.paste(strip_top, (self._border, 0))
            tmp.paste(strip_bottom, (self._border, tmp.height - self._border))
            tmp.paste(strip_left, (0, self._border))
            tmp.paste(strip_right, (tmp.width - self._border, self._border))
        else:
            tmp = image

        # Write the image directly to graphics memory in the allocated space
        self._texture.write(tmp.tobytes(), 0, viewport=viewport)

    def _remove_texture_by_identifiers(self, atlas_name: str, hash: str):
        """
        Called by the finalizer to remove a texture by its identifiers.
        This should never be called directly.

        Args:
            atlas_name: The name of the texture in the atlas
            hash: The hash of the image data
        """
        # LOG.info("Removing texture: %s", atlas_name)
        # print("Removing texture:", atlas_name)

        # Remove the unique texture if ref counter reaches 0
        if self._unique_texture_ref_count.dec_ref_by_atlas_name(atlas_name) == 0:
            # Remove the unique texture key to signal we don't have any more
            refs = self._unique_textures[atlas_name]
            if len(refs) == 0:
                del self._unique_textures[atlas_name]

            # Reclaim region and uv slot
            del self._texture_regions[atlas_name]
            self._texture_uvs.free_slot_by_name(atlas_name)

        # Remove the image if ref counter reaches 0
        if self._image_ref_count.dec_ref_by_hash(hash) == 0:
            # May have been removed by GC
            try:
                del self._images[hash]
            except KeyError:
                pass

            del self._image_regions[hash]

            # Reclaim the image uv slot
            self._image_uvs.free_slot_by_name(hash)

        self._textures_removed += 1

    def update_texture_image(self, texture: Texture):
        """
        Updates the internal image of a texture in the atlas texture.
        The new image needs to be the exact same size as the original
        one meaning the texture already need to exist in the atlas.

        This can be used in cases were the image is manipulated in some way
        and we need a quick way to sync these changes to graphics memory.
        This operation is fairly expensive, but still orders of magnitude
        faster than removing the old texture, adding the new one and
        re-building the entire atlas.

        Args:
            texture: The texture to update
        """
        region = self._image_regions[texture.image_data.hash]
        region.verify_image_size(texture.image_data)
        viewport = (
            region.x,
            region.y,
            region.width,
            region.height,
        )
        self._texture.write(texture.image.tobytes(), 0, viewport=viewport)

    def get_image_region_info(self, hash: str) -> AtlasRegion:
        """
        Get the region info for and image by has

        Args:
            hash: The hash of the image
        """
        return self._image_regions[hash]

    def get_texture_region_info(self, atlas_name: str) -> AtlasRegion:
        """
        Get the region info for a texture by atlas name

        Args:
            atlas_name: The name of the texture in the atlas
        """
        return self._texture_regions[atlas_name]

    def get_texture_id(self, texture: Texture) -> int:
        """
        Get the internal id for a Texture in the atlas

        Args:
            texture: The texture to get.
        """
        return self._texture_uvs.get_slot_or_raise(texture.atlas_name)

    def has_texture(self, texture: Texture) -> bool:
        """
        Check if a texture is already in the atlas.

        Args:
            texture: The texture to check
        """
        return texture in self._textures

    def has_unique_texture(self, texture: Texture) -> bool:
        """
        Check if the atlas already have a texture with the
        same image data and vertex order

        Args:
            texture: The texture to check
        """
        return texture.atlas_name in self._unique_textures

    def has_image(self, image_data: ImageData) -> bool:
        """
        Check if an image is already in the atlas

        Args:
            image_data: The image data to check
        """
        return image_data.hash in self._images

    def resize(self, size: tuple[int, int], force=False) -> None:
        """
        Resize the atlas.

        This will re-allocate all the images in the atlas to better fit
        the new size. Pixel data will be copied from the old atlas to the
        new one on the gpu meaning it will also persist anything that
        was rendered to the atlas.

        A failed resize will result in an AllocatorException. Unless the
        atlas is resized again to a working size the atlas will be in an
        undefined state.

        Args:
            size:
                The new size
            force:
                Force a resize even if the size is the same
        """
        # LOG.info("[%s] Resizing atlas from %s to %s", id(self), self._size, size)
        # print("Resizing atlas from", self._size, "to", size)

        # Only resize if the size actually changed
        if size == self._size and not force:
            return

        self._check_size(size)
        # resize_start = time.perf_counter()

        # Keep a reference to the old atlas texture so we can copy it into the new one
        atlas_texture_old = self._texture
        atlas_texture_old.filter = self._ctx.NEAREST, self._ctx.NEAREST
        self._size = size

        # Create new image uv data temporarily keeping the old one
        self._image_uvs.write_to_texture()
        image_uvs_old = self._image_uvs
        self._image_uvs = image_uvs_old.clone_with_slots()

        # Create new atlas texture and framebuffer
        self._texture = self._ctx.texture(size, components=4)
        self._fbo = self._ctx.framebuffer(color_attachments=[self._texture])

        # Store old images and textures before clearing the atlas
        images = list(self._images.values())
        textures = self.unique_textures

        # Clear the regions and allocator.
        self._allocator = Allocator(*self._size)
        # NOTE: We keep the image_regions and texture_regions in case the resize fails

        # Re-allocate the images
        for image in sorted(images, key=lambda x: x.height):
            self._allocate_image(image)
        self._image_uvs.write_to_texture()

        # Update the texture regions. We need to copy the image regions
        # and re-apply the transforms on each texture
        for texture in textures:
            self._allocate_texture(texture)
        self._texture_uvs.write_to_texture()

        # Bind textures for atlas copy shader
        atlas_texture_old.use(0)
        self._texture.use(1)
        image_uvs_old.texture.use(2)
        self._image_uvs.texture.use(3)
        self._ctx.atlas_resize_program["border"] = float(self._border)
        self._ctx.atlas_resize_program["projection"] = Mat4.orthogonal_projection(
            0,
            self.width,
            self.height,
            0,
            -100,
            100,
        )

        # Render the old atlas into the new one. This means we actually move
        # all the textures around from the old to the new position.
        with self._fbo.activate():
            # Ensure no context flags are enabled
            with self._ctx.enabled_only():
                self._ctx.geometry_empty.render(
                    self._ctx.atlas_resize_program,
                    mode=self._ctx.POINTS,
                    vertices=self.max_width,
                )

        # duration = time.perf_counter() - resize_start
        # LOG.info("[%s] Atlas resize took %s seconds", id(self), duration)

    def rebuild(self) -> None:
        """
        Rebuild the underlying atlas texture.

        This method also tries to organize the textures more efficiently ordering them by size.
        The texture ids will persist so the sprite list doesn't need to be rebuilt.
        """
        # LOG.info("Rebuilding atlas")

        # Hold a reference to the old textures
        textures = self.textures

        self._image_ref_count.clear()
        self._unique_texture_ref_count.clear()

        # Clear the atlas but keep the uv slot mapping
        self._fbo.clear()

        self._textures.clear()
        self._unique_textures.clear()
        self._images.clear()

        self._image_regions.clear()
        self._texture_regions.clear()
        self._allocator = Allocator(*self._size)

        # Add textures back sorted by height to potentially make more room
        for texture in sorted(textures, key=lambda x: x.image.size[1]):
            self._add(texture, create_finalizer=False)

    def use_uv_texture(self, unit: int = 0) -> None:
        """
        Bind the texture coordinate texture to a channel.
        In addition this method writes the texture
        coordinate to the texture if the data is stale.
        This is to avoid a full update every time a texture
        is added to the atlas.

        Args:
            unit: The texture unit to bind the uv texture
        """
        # Sync the texture coordinates to the texture if dirty
        self._image_uvs.write_to_texture()
        self._texture_uvs.write_to_texture()

        self._texture_uvs.texture.use(unit)

    @contextlib.contextmanager
    def render_into(
        self,
        texture: Texture,
        projection: tuple[float, float, float, float] | None = None,
    ):
        """
        Render directly into a sub-section of the atlas.
        The sub-section is defined by the already allocated space
        of the texture supplied in this method.

        By default the projection will be set to match the texture area size
        were `0, 0` is the lower left corner and `width, height` (of texture)
        is the upper right corner.

        This method should should be used with the ``with`` statement::

            with atlas.render_into(texture):
                # Draw commands here

            # Specify projection
            with atlas.render_into(texture, projection=(0, 100, 0, 100))
                # Draw geometry

        Args:
            texture:
                The texture area to render into
            projection:
                The ortho projection to render with. This parameter can be
                left blank if no projection changes are needed.
                The tuple values are: (left, right, button, top)
        """
        region = self._texture_regions[texture.atlas_name]
        prev_camera = self.ctx.current_camera

        # Use provided projection or default
        projection = projection or (0, region.width, 0, region.height)
        # Flip the top and bottom because we need to render things upside down
        projection = projection[0], projection[1], projection[3], projection[2]

        static_camera = static_from_raw_orthographic(
            projection,
            -1,
            1,  # near, far planes
            1.0,  # zoom
        )

        with self._fbo.activate() as fbo:
            try:
                static_camera.use()
                fbo.viewport = region.x, region.y, region.width, region.height
                yield fbo
            finally:
                fbo.viewport = 0, 0, *self._fbo.size
        prev_camera.use()

    def read_texture_image_from_atlas(self, texture: Texture) -> Image.Image:
        """
        Read the pixel data for a texture directly from the atlas texture on the GPU.
        The contents of this image can be altered by rendering into the atlas and
        is useful in situations were you need the updated pixel data on the python side.

        Args:
            texture: The texture to get the image for
        Returns:
            A pillow image containing the pixel data in the atlas
        """
        region = self.get_image_region_info(texture.image_data.hash)
        viewport = (
            region.x,
            region.y,
            region.width,
            region.height,
        )
        data = self.fbo.read(viewport=viewport, components=4)
        return Image.frombytes("RGBA", (region.width, region.height), data)

    def update_texture_image_from_atlas(self, texture: Texture) -> None:
        """
        Update the Arcade Texture's internal image with the pixel data content
        from the atlas texture on the GPU. This can be useful if you render
        into the atlas and need to update the texture with the new pixel data.

        Args:
            texture: The texture to update
        """
        texture.image_data.image = self.read_texture_image_from_atlas(texture)

    def to_image(
        self,
        flip: bool = False,
        components: int = 4,
        draw_borders: bool = False,
        border_color: tuple[int, int, int] = (255, 0, 0),
    ) -> Image.Image:
        """
        Convert the atlas to a Pillow image.

        Borders can also be drawn into the image to visualize the
        regions of the atlas.

        Args:
            flip:
                Flip the image horizontally
            components:
                Number of components. (3 = RGB, 4 = RGBA)
            draw_borders:
                Draw region borders into image
            border_color:
                RGB color of the borders
        Returns:
            A pillow image containing the atlas texture
        """
        if components not in (3, 4):
            raise ValueError(f"Components must be 3 or 4, not {components}")

        mode = "RGBA"[:components]

        image = Image.frombytes(
            mode,
            self._texture.size,
            bytes(self._fbo.read(components=components)),
        )

        if draw_borders:
            draw = ImageDraw.Draw(image)
            for rg in self._image_regions.values():
                p1 = rg.x, rg.y
                p2 = rg.x + rg.width - 1, rg.y + rg.height - 1
                draw.rectangle((p1, p2), outline=border_color, width=1)

        if flip:
            image = image.transpose(Image.Transpose.FLIP_TOP_BOTTOM)

        return image

    def show(
        self,
        flip: bool = False,
        components: int = 4,
        draw_borders: bool = False,
        border_color: tuple[int, int, int] = (255, 0, 0),
    ) -> None:
        """
        Show the texture atlas using Pillow.

        Borders can also be drawn into the image to visualize the
        regions of the atlas.

        Args:
            flip:
                Flip the image horizontally
            components:
                Number of components. (3 = RGB, 4 = RGBA)
            draw_borders:
                Draw region borders into image
            border_color:
                RGB color of the borders
        """
        self.to_image(
            flip=flip,
            components=components,
            draw_borders=draw_borders,
            border_color=border_color,
        ).show()

    def save(
        self,
        path: str | Path,
        flip: bool = False,
        components: int = 4,
        draw_borders: bool = False,
        border_color: tuple[int, int, int] = (255, 0, 0),
    ) -> None:
        """
        Save the texture atlas to a png.

        Borders can also be drawn into the image to visualize the
        regions of the atlas.

        Args:
            path:
                The path to save the atlas on disk
            flip:
                Flip the image horizontally
            components:
                Number of components. (3 = RGB, 4 = RGBA)
            draw_borders:
                Draw region borders into image
            border_color:
                RGB color of the borders
        """
        self.to_image(
            flip=flip,
            components=components,
            draw_borders=draw_borders,
            border_color=border_color,
        ).save(path, format="png")

    def _check_size(self, size: tuple[int, int]) -> None:
        """Check it the atlas exceeds the hardware limitations"""
        if size[0] > self._max_size[0] or size[1] > self._max_size[1]:
            raise Exception(
                "Attempting to create or resize an atlas to "
                f"{size} past its maximum size of {self._max_size}"
            )



"""
Original screenshot utilities from Arcade 2.

These functions are flawed because they only read from the screen.
"""

import PIL.Image
import PIL.ImageOps

from arcade.window_commands import get_window


def get_pixel(x: int, y: int, components: int = 3) -> tuple[int, ...]:
    """
    Given an x, y, will return a color value of that point.

    Args:
        x: x location
        y: y location
        components: Number of components to fetch. By default we fetch 3
            3 components (RGB). 4 components would be RGBA.
    """
    # The window may be 'scaled' on hi-res displays. Particularly Macs. OpenGL
    # won't account for this, so we need to.
    window = get_window()
    ctx = window.ctx

    pixel_ratio = window.get_pixel_ratio()
    x = int(pixel_ratio * x)
    y = int(pixel_ratio * y)

    data = ctx.screen.read(viewport=(x, y, 1, 1), components=components)
    return tuple(data)  # bytes gets converted to ints in the tuple creation


def get_image(
    x: int = 0,
    y: int = 0,
    width: int | None = None,
    height: int | None = None,
    components: int = 4,
) -> PIL.Image.Image:
    """
    Get an image from the screen.

    Example::

        # Create and image of the entire screen and save it to a file
        image = arcade.get_image()
        image.save('screenshot.png')

    Args:
        x: Start (left) x location
        y: Start (bottom) y location
        width: Width of image. Leave blank for grabbing the 'rest' of the image
        height: Height of image. Leave blank for grabbing the 'rest' of the image
        components: Number of components to fetch. By default we fetch 4 (4=RGBA, 3=RGB)
    """
    window = get_window()
    ctx = window.ctx

    if components not in (3, 4):
        raise ValueError("components must be 3 or 4")

    pixel_ratio = window.get_pixel_ratio()
    x = int(pixel_ratio * x)
    y = int(pixel_ratio * y)

    if width is None:
        width = window.width - x
    if height is None:
        height = window.height - y

    width = int(pixel_ratio * width)
    height = int(pixel_ratio * height)

    data = ctx.screen.read(viewport=(x, y, width, height), components=components)
    image = PIL.Image.frombytes("RGBA" if components == 4 else "RGB", (width, height), data)
    return PIL.ImageOps.flip(image)



"""
Sprite Explosion

Simple program to show creating explosions with sprites.
There are more performant ways to do this, but for simple games this
is a good way to get started.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_explosion_particles
"""
import random
import math
import arcade

SPRITE_SCALING_PLAYER = 0.5
SPRITE_SCALING_COIN = 0.3
SPRITE_SCALING_LASER = 0.8
ENEMY_COUNT = 50

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Explosion Example"

BULLET_SPEED = 5

# --- Explosion Particles Related

# How fast the particle will accelerate down. Make 0 if not desired
PARTICLE_GRAVITY = 0.05

# How fast to fade the particle
PARTICLE_FADE_RATE = 8

# How fast the particle moves. Range is from 2.5 <--> 5 with 2.5 and 2.5 set.
PARTICLE_MIN_SPEED = 2.5
PARTICLE_SPEED_RANGE = 2.5

# How many particles per explosion
PARTICLE_COUNT = 20

# How big the particle
PARTICLE_RADIUS = 3

# Possible particle colors
PARTICLE_COLORS = [arcade.color.ALIZARIN_CRIMSON,
                   arcade.color.COQUELICOT,
                   arcade.color.LAVA,
                   arcade.color.KU_CRIMSON,
                   arcade.color.DARK_TANGERINE]

# Chance we'll flip the texture to white and make it 'sparkle'
PARTICLE_SPARKLE_CHANCE = 0.02

# --- Smoke
# Note: Adding smoke trails makes for a lot of sprites and can slow things
# down. If you want a lot, it will be necessary to move processing to GPU
# using transform feedback. If to slow, just get rid of smoke.

# Start scale of smoke, and how fast is scales up
SMOKE_START_SCALE = 0.25
SMOKE_EXPANSION_RATE = 0.03

# Rate smoke fades, and rises
SMOKE_FADE_RATE = 7
SMOKE_RISE_RATE = 0.5

# Chance we leave smoke trail
SMOKE_CHANCE = 0.25


class Smoke(arcade.SpriteCircle):
    """Particle with smoke like behavior."""
    def __init__(self, size):
        super().__init__(size, arcade.color.LIGHT_GRAY, soft=True)
        self.change_y = SMOKE_RISE_RATE
        self.scale = SMOKE_START_SCALE

    def update(self, delta_time: float = 1/60):
        """Update this particle"""
        # Take delta_time into account
        time_step = 60 * delta_time

        if self.alpha <= PARTICLE_FADE_RATE:
            # Remove faded out particles
            self.remove_from_sprite_lists()
        else:
            # Update values
            self.alpha -= int(SMOKE_FADE_RATE * time_step)
            self.center_x += self.change_x * time_step
            self.center_y += self.change_y * time_step
            self.add_scale(SMOKE_EXPANSION_RATE * time_step)


class Particle(arcade.SpriteCircle):
    """ Explosion particle"""
    def __init__(self):
        """
        Simple particle sprite based on circle sprite.
        """
        # Make the particle
        super().__init__(PARTICLE_RADIUS, random.choice(PARTICLE_COLORS))

        # Set direction/speed
        speed = random.random() * PARTICLE_SPEED_RANGE + PARTICLE_MIN_SPEED
        direction = random.randrange(360)
        self.change_x = math.sin(math.radians(direction)) * speed
        self.change_y = math.cos(math.radians(direction)) * speed

    def update(self, delta_time: float = 1 / 60):
        """Update the particle"""
        # Take delta_time into account
        time_step = 60 * delta_time

        if self.alpha == 0:
            # Faded out, remove
            self.remove_from_sprite_lists()
        else:
            # Gradually fade out the particle. Don't go below 0
            self.alpha = max(0, self.alpha - PARTICLE_FADE_RATE)
            # Move the particle
            self.center_x += self.change_x * time_step
            self.center_y += self.change_y * time_step
            self.change_y -= PARTICLE_GRAVITY * time_step

            # Should we sparkle this?
            if random.random() <= PARTICLE_SPARKLE_CHANCE:
                self.alpha = 255
                self.color = arcade.color.WHITE

            # Leave a smoke particle?
            if random.random() <= SMOKE_CHANCE:
                smoke = Smoke(5)
                smoke.position = self.position
                # Add a smoke particle to the spritelist this sprite is in
                self.sprite_lists[0].append(smoke)


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """ Initializer """
        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.player_list = arcade.SpriteList()
        self.enemy_list = arcade.SpriteList()
        self.bullet_list = arcade.SpriteList()
        self.explosions_list = arcade.SpriteList()

        # Set up the player info. Image from kenney.nl
        self.player_sprite = arcade.Sprite(
            ":resources:images/space_shooter/playerShip2_orange.png",
            scale=SPRITE_SCALING_PLAYER,
        )
        self.player_sprite.center_x = 50
        self.player_sprite.center_y = 70
        self.player_list.append(self.player_sprite)

        self.score = 0

        # Don't show the mouse cursor
        self.window.set_mouse_visible(False)

        # Load sounds. Sounds from kenney.nl
        self.gun_sound = arcade.sound.load_sound(":resources:sounds/laser2.wav")
        self.hit_sound = arcade.sound.load_sound(":resources:sounds/explosion2.wav")

        self.background_color = arcade.color.BLACK
        self.score_display = arcade.Text("", 10, 20, arcade.color.WHITE, 14)

        self.spawn_enemies()

    def reset(self):
        """Restart the game"""
        # Reset score
        self.score = 0

        self.enemy_list.clear()
        self.bullet_list.clear()
        self.explosions_list.clear()

        self.spawn_enemies()

    def spawn_enemies(self):
        # Spawn enemies
        for index in range(ENEMY_COUNT):
            # Create the coin instance. Image from kenney.nl
            enemy = arcade.Sprite(
                ":resources:images/space_shooter/playerShip1_green.png",
                scale=SPRITE_SCALING_COIN,
                angle=180,
                center_x=random.randrange(25, WINDOW_WIDTH - 25),
                center_y=random.randrange(150, WINDOW_HEIGHT)
            )
            # Add the ship to the lists
            self.enemy_list.append(enemy)

    def on_draw(self):
        """
        Render the screen.
        """
        # This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites.
        self.enemy_list.draw()
        self.bullet_list.draw()
        self.player_list.draw()
        self.explosions_list.draw()

        # Render the text
        self.score_display.text = f"Score: {self.score}"
        self.score_display.draw()

    def on_mouse_motion(self, x, y, dx, dy):
        """
        Called whenever the mouse moves.
        """
        self.player_sprite.center_x = x

    def on_mouse_press(self, x, y, button, modifiers):
        """
        Called whenever the mouse button is clicked.
        """
        # Gunshot sound
        arcade.sound.play_sound(self.gun_sound)

        # Create a bullet
        bullet = arcade.Sprite(
            ":resources:images/space_shooter/laserBlue01.png",
            scale=SPRITE_SCALING_LASER,
        )

        # The image points to the right, and we want it to point up. So
        # rotate it.
        bullet.angle = 270

        # Give it a speed
        bullet.change_y = BULLET_SPEED

        # Position the bullet
        bullet.center_x = self.player_sprite.center_x
        bullet.bottom = self.player_sprite.top

        # Add the bullet to the appropriate lists
        self.bullet_list.append(bullet)

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol == arcade.key.R:
            self.reset()
        # Close the window
        elif symbol == arcade.key.ESCAPE:
            self.window.close()

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on bullet sprites
        self.bullet_list.update(delta_time)
        self.explosions_list.update(delta_time)

        # Loop through each bullet
        for bullet in self.bullet_list:

            # Check this bullet to see if it hit a coin
            hit_list = arcade.check_for_collision_with_list(bullet, self.enemy_list)

            # If it did...
            if len(hit_list) > 0:

                # Get rid of the bullet
                bullet.remove_from_sprite_lists()

            # For every coin we hit, add to the score and remove the coin
            for coin in hit_list:
                # Make an explosion
                for i in range(PARTICLE_COUNT):
                    particle = Particle()
                    particle.position = coin.position
                    self.explosions_list.append(particle)

                smoke = Smoke(50)
                smoke.position = coin.position
                self.explosions_list.append(smoke)

                coin.remove_from_sprite_lists()
                self.score += 1

                # Hit Sound
                arcade.sound.play_sound(self.hit_sound)

            # If the bullet flies off-screen, remove it.
            if bullet.bottom > WINDOW_HEIGHT:
                bullet.remove_from_sprite_lists()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

import warnings
from abc import ABCMeta, abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass

from pyglet.event import EVENT_HANDLED, EVENT_UNHANDLED
from typing_extensions import override

import arcade
from arcade import Texture, uicolor
from arcade.gui import (
    NinePatchTexture,
    Surface,
    UIEvent,
    UIInteractiveWidget,
    UIMouseDragEvent,
    UIOnClickEvent,
)
from arcade.gui.events import UIOnChangeEvent
from arcade.gui.property import Property, bind
from arcade.gui.style import UIStyleBase, UIStyledWidget
from arcade.types import RGBA255


class UIBaseSlider(UIInteractiveWidget, metaclass=ABCMeta):
    """Base class for sliders.

    A slider contains of a horizontal track and a thumb.
    The thumb can be moved along the track to set the value of the slider.

    Use the `on_change` event to get notified about value changes.

    Subclasses should implement the `_render_track` and `_render_thumb` methods.

    Args:

        value: Current value of the curosr of the slider.
        min_value: Minimum value of the slider.
        max_value: Maximum value of the slider.
        x: x coordinate of bottom left.
        y: y coordinate of bottom left.
        width: Width of the slider.
        height: Height of the slider.
        size_hint: Size hint of the slider.
        size_hint_min: Minimum size hint of the slider.
        size_hint_max: Maximum size hint of the slider.
        style: Used to style the slider for different states.
        step: Smallest change the slider value can move by.
        **kwargs: Passed to UIInteractiveWidget.

    """

    value = Property(0.0)

    def __init__(
        self,
        *,
        value: float = 0,
        min_value: float = 0,
        max_value: float = 100,
        x: float = 0,
        y: float = 0,
        width: float = 300,
        height: float = 20,
        size_hint=None,
        size_hint_min=None,
        size_hint_max=None,
        style: Mapping[str, UISliderStyle] | None = None,
        step: float | None = None,
        **kwargs,
    ):
        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            style=style or UISlider.DEFAULT_STYLE,
            **kwargs,
        )

        self.step = step
        self.value = value
        self.min_value = min_value
        self.max_value = max_value

        self._cursor_width = self.height // 3

        # trigger render on value changes
        bind(self, "value", self.trigger_full_render)
        bind(self, "value", self._ensure_step)
        bind(self, "hovered", self.trigger_render)
        bind(self, "pressed", self.trigger_render)
        bind(self, "disabled", self.trigger_render)

        self.register_event_type("on_change")

    def _ensure_step(self):
        """Ensure that the step is applied."""
        if self.step is not None:
            # this will trigger the change once again
            # only option to prevent this would be to make `value` a property,
            # which might break code of users
            self.value = self._apply_step(self.value)

    def _apply_step(self, value: float):
        if self.step:
            inverse = 1 / self.step
            return round(value * inverse) / inverse

        return value

    def _set_value(self, value: float):
        # TODO changing the value itself should trigger `on_change` event
        # current problem is, that the property does not pass the old value to listeners
        if value < self.min_value:
            value = self.min_value
        elif value > self.max_value:
            value = self.max_value

        if self.value == value:
            return

        old_value = self.value
        self.value = value
        self.dispatch_event("on_change", UIOnChangeEvent(self, old_value, self.value))

    def _x_for_value(self, value: float):
        """Provides the x coordinate for the given value."""

        x = self.content_rect.left
        val = (value - self.min_value) / (self.max_value - self.min_value)
        return x + self._cursor_width + val * (self.content_width - 2 * self._cursor_width)

    @property
    def norm_value(self):
        """Normalized value between 0.0 and 1.0"""
        return (self.value - self.min_value) / self.max_value

    @norm_value.setter
    def norm_value(self, value):
        """Normalized value between 0.0 and 1.0"""
        new_value = min(value * (self.max_value - self.min_value) + self.min_value, self.max_value)
        self._set_value(new_value)

    @property
    def _thumb_x(self):
        """Returns the current x coordinate of the thumb."""
        return self._x_for_value(self.value)

    @_thumb_x.setter
    def _thumb_x(self, nx):
        """Set thumb x coordinate and update the value."""
        rect = self.content_rect

        x = min(rect.right - self._cursor_width, max(nx, rect.left + self._cursor_width))
        if self.width == 0:
            self.norm_value = 0
        else:
            self.norm_value = (x - rect.left - self._cursor_width) / float(
                self.content_width - 2 * self._cursor_width
            )

    @override
    def do_render(self, surface: Surface):
        """Render the slider, including track and thumb."""
        self.prepare_render(surface)
        self._render_track(surface)
        self._render_steps(surface)
        self._render_thumb(surface)

    @abstractmethod
    def _render_track(self, surface: Surface):
        """Render the track of the slider.

        This method should be implemented in a slider implementation.

        Track should stay within self.content_rect.

        Args:
                surface: Surface to render on.
        """
        pass

    @abstractmethod
    def _render_steps(self, surface: Surface):
        """Render the steps of the slider track.

        This method should be implemented in a slider implementation.

        Steps should stay within self.content_rect.

        Args:
                surface: Surface to render on.
        """
        pass

    @abstractmethod
    def _render_thumb(self, surface: Surface):
        """Render the thumb of the slider.

        This method should be implemented in a slider implementation.

        Thumb should stay within self.content_rect.
        x coordinate of the thumb should be self._thumb_x.

        Args:
            surface: Surface to render on.
        """
        pass

    @override
    def on_event(self, event: UIEvent) -> bool | None:
        """
        Args:
            event: Event to handle.

        Returns: True if event was handled, False otherwise.

        """
        if self.disabled:
            return EVENT_UNHANDLED

        if super().on_event(event):
            return EVENT_HANDLED

        if isinstance(event, UIMouseDragEvent):
            if self.pressed:
                self._thumb_x = event.x

                return EVENT_HANDLED

        return EVENT_UNHANDLED

    @override
    def on_click(self, event: UIOnClickEvent):
        """Handle click events to set the value of the slider.

        A new value is calculated based on the click position and the slider's width and
        the `on_change` event is dispatched.

        Args:
            event: Click event.
        """
        old_value = self.value
        self._thumb_x = event.x
        self.dispatch_event("on_change", UIOnChangeEvent(self, old_value, self.value))

    def on_change(self, event: UIOnChangeEvent):
        """To be implemented by the user, triggered when the cursor's value is changed.

        Args:
            event: Event containing the old and new value of the cursor.
        """
        pass


@dataclass
class UISliderStyle(UIStyleBase):
    """Used to style the slider for different states. Below is its use case.

    .. code:: py

        button = UITextureButton(style={"normal": UITextureButton.UIStyle(...),})

    Args:
        bg: Background color.
        border: Border color.
        border_width: Width of the border.
        filled_track: Color of the filled track.
        filled_step: Color of the step in filled area.
        unfilled_track: Color of the unfilled track.
        unfilled_step: Color of the step in unfilled area.
    """

    bg: RGBA255 = uicolor.WHITE_SILVER
    border: RGBA255 = uicolor.DARK_BLUE_MIDNIGHT_BLUE
    border_width: int = 2
    filled_track: RGBA255 = uicolor.DARK_BLUE_MIDNIGHT_BLUE
    filled_step: RGBA255 | None = uicolor.BLUE_PETER_RIVER
    unfilled_track: RGBA255 = uicolor.WHITE_SILVER
    unfilled_step: RGBA255 | None = uicolor.BLUE_PETER_RIVER


class UISlider(UIStyledWidget[UISliderStyle], UIBaseSlider):
    """A simple slider.

    A slider consists of a horizontal track and a thumb.
    The thumb can be moved along the track to set the value of the slider.

    Use the `on_change` event to get notified about value changes.

    There are four states of the UISlider i.e. normal, hovered, pressed and disabled.

    Args:
        value: Current value of the cursor of the slider.
        min_value: Minimum value of the slider.
        max_value: Maximum value of the slider.
        x: x coordinate of bottom left.
        y: y coordinate of bottom left.
        width: Width of the slider.
        height: Height of the slider.
        style: Used to style the slider for different states.
        step: Smallest change the slider value can move by.
    """

    UIStyle = UISliderStyle

    DEFAULT_STYLE = {
        "normal": UIStyle(),
        "hover": UIStyle(
            border=uicolor.BLUE_PETER_RIVER,
            border_width=2,
            filled_track=uicolor.BLUE_PETER_RIVER,
            filled_step=uicolor.DARK_BLUE_MIDNIGHT_BLUE,
        ),
        "press": UIStyle(
            bg=uicolor.BLUE_PETER_RIVER,
            border=uicolor.DARK_BLUE_WET_ASPHALT,
            border_width=3,
            filled_track=uicolor.BLUE_PETER_RIVER,
            filled_step=uicolor.DARK_BLUE_MIDNIGHT_BLUE,
        ),
        "disabled": UIStyle(
            bg=uicolor.WHITE_SILVER,
            border_width=1,
            filled_track=uicolor.GRAY_ASBESTOS,
            unfilled_track=uicolor.WHITE_SILVER,
        ),
    }

    NO_STEP_STYLE = {
        "normal": UIStyle(
            filled_step=None,
            unfilled_step=None,
        ),
        "hover": UIStyle(
            border=uicolor.BLUE_PETER_RIVER,
            border_width=2,
            filled_track=uicolor.BLUE_PETER_RIVER,
            filled_step=None,
            unfilled_step=None,
        ),
        "press": UIStyle(
            bg=uicolor.BLUE_PETER_RIVER,
            border=uicolor.DARK_BLUE_WET_ASPHALT,
            border_width=3,
            filled_track=uicolor.BLUE_PETER_RIVER,
            filled_step=None,
            unfilled_step=None,
        ),
        "disabled": UIStyle(
            bg=uicolor.WHITE_SILVER,
            border_width=1,
            filled_track=uicolor.GRAY_ASBESTOS,
            unfilled_track=uicolor.WHITE_SILVER,
            filled_step=None,
            unfilled_step=None,
        ),
    }
    """Removing the step colors from the style.
    So sliders with a step value do not show the steps visually."""

    def __init__(
        self,
        *,
        value: float = 0,
        min_value: float = 0,
        max_value: float = 100,
        x: float = 0,
        y: float = 0,
        width: float = 300,
        height: float = 25,
        size_hint=None,
        size_hint_min=None,
        size_hint_max=None,
        style: dict[str, UISliderStyle] | None = None,
        step: float | None = None,
        **kwargs,
    ):
        super().__init__(
            value=value,
            min_value=min_value,
            max_value=max_value,
            x=x,
            y=y,
            width=width,
            height=height,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            style=style or UISlider.DEFAULT_STYLE,
            step=step,
            **kwargs,
        )

    @override
    def get_current_state(self) -> str:
        """Get the current state of the slider.

        Returns:
            ""normal"", ""hover"", ""press"" or ""disabled"".
        """
        if self.disabled:
            return "disabled"
        elif self.pressed:
            return "press"
        elif self.hovered:
            return "hover"
        else:
            return "normal"

    @override
    def _render_track(self, surface: Surface):
        style = self.get_current_style()
        if style is None:
            warnings.warn(f"No style found for state {self.get_current_state()}", UserWarning)
            return

        bg_slider_color = style.get("unfilled_track", UISlider.UIStyle.unfilled_track)
        fg_slider_color = style.get("filled_track", UISlider.UIStyle.filled_track)

        slider_height = self.content_height // 3

        slider_left_x = self._x_for_value(self.min_value)
        slider_right_x = self._x_for_value(self.max_value)
        cursor_center_x = self._thumb_x

        slider_bottom = (self.content_height - slider_height) // 2

        arcade.draw_lbwh_rectangle_filled(
            slider_left_x - self.content_rect.left,
            slider_bottom,
            slider_right_x - slider_left_x,
            slider_height,
            bg_slider_color,
        )
        arcade.draw_lbwh_rectangle_filled(
            slider_left_x - self.content_rect.left,
            slider_bottom,
            cursor_center_x - slider_left_x,
            slider_height,
            fg_slider_color,
        )

    @override
    def _render_steps(self, surface: Surface):
        if not self.step:
            return

        style = self.get_current_style()
        if style is None:
            warnings.warn(f"No style found for state {self.get_current_state()}", UserWarning)
            return

        unfilled_steps = style.get("unfilled_step", UISlider.UIStyle.unfilled_step)
        filled_steps = style.get("filled_step", UISlider.UIStyle.filled_step)

        def float_range(start, stop, step):
            while start < stop:
                yield start
                start += step
            yield stop

        steps = list(float_range(self.min_value, self.max_value, self.step))

        for v in steps:
            step_x = self._x_for_value(v) - self.content_rect.left
            step_color = filled_steps if v <= self.value else unfilled_steps

            if step_color:
                # bigger circle for first and last step
                circle_size = self._cursor_width // 4
                if v in (steps[0], steps[-1]):
                    circle_size = self._cursor_width // 2

                arcade.draw_circle_filled(
                    step_x,
                    self.content_height // 2,
                    circle_size,
                    step_color,
                    num_segments=8,
                )

    @override
    def _render_thumb(self, surface: Surface):
        style = self.get_current_style()
        if style is None:
            warnings.warn(f"No style found for state {self.get_current_state()}", UserWarning)
            return

        border_width = style.get("border_width", UISlider.UIStyle.border_width)
        cursor_color = style.get("bg", UISlider.UIStyle.bg)
        cursor_outline_color = style.get("border", UISlider.UIStyle.border)

        cursor_radius = self._cursor_width
        cursor_center_x = self._thumb_x
        slider_center_y = self.content_height // 2

        rel_cursor_x = cursor_center_x - self.content_rect.left
        arcade.draw_circle_filled(rel_cursor_x, slider_center_y, cursor_radius, cursor_color)
        arcade.draw_circle_filled(
            rel_cursor_x, slider_center_y, cursor_radius // 4, cursor_outline_color
        )
        arcade.draw_circle_outline(
            rel_cursor_x,
            slider_center_y,
            cursor_radius,
            cursor_outline_color,
            border_width,
        )


class UITextureSlider(UISlider):
    """A custom slider subclass which supports textures.

    You can copy this as-is into your own project, or you can modify
    the class to have more features as needed.

    Args:
        track_texture: Texture for the track, should be a NinePatchTexture.
        thumb_texture: Texture for the thumb.
        style: Used to style the slider for different states.
        **kwargs: Passed to UISlider.
    """

    def __init__(
        self,
        track_texture: Texture | NinePatchTexture,
        thumb_texture: Texture | NinePatchTexture,
        style=None,
        **kwargs,
    ):
        self._track_tex = track_texture
        self._thumb_tex = thumb_texture

        super().__init__(style=style or UISlider.DEFAULT_STYLE, **kwargs)

    @override
    def _render_track(self, surface: Surface):
        style = self.get_current_style()
        if style is None:
            warnings.warn(f"No style found for state {self.get_current_state()}", UserWarning)
            return

        surface.draw_texture(0, 0, self.width, self.height, self._track_tex)

        # TODO accept these as constructor params
        slider_height = self.height // 4
        slider_left_x = self._x_for_value(self.min_value)
        cursor_center_x = self._thumb_x

        slider_bottom = (self.height - slider_height) // 2

        # slider
        if style.filled_track:
            arcade.draw_lbwh_rectangle_filled(
                slider_left_x - self.left,
                slider_bottom,
                cursor_center_x - slider_left_x,
                slider_height,
                style.filled_track,
            )

    @override
    def _render_thumb(self, surface: Surface):
        cursor_center_x = self._thumb_x
        rel_cursor_x = cursor_center_x - self.left
        surface.draw_texture(
            x=rel_cursor_x - self._thumb_tex.width // 4 + 2,
            y=0,
            width=self._thumb_tex.width // 2,
            height=self.height,
            tex=self._thumb_tex,
        )



"""
Shows how we can use shaders using existing spritelist data.

This examples renders a line between the player position
and nearby sprites when they are within a certain distance.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.spritelist_interaction_visualize_dist
"""

import random
import arcade

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
NUM_COINS = 1000
INTERACTION_RADIUS = 250


class SpriteListInteraction(arcade.Window):

    def __init__(self):
        super().__init__(WINDOW_WIDTH, WINDOW_HEIGHT, "SpriteList Interaction")

        # Generate some random coins
        self.coins = arcade.SpriteList()
        for _ in range(NUM_COINS):
            self.coins.append(
                arcade.Sprite(
                    ":resources:images/items/coinGold.png",
                    center_x=random.randint(0, WINDOW_WIDTH),
                    center_y=random.randint(0, WINDOW_HEIGHT),
                    scale=0.25,
                )
            )

        # Player
        self.player = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=0.5,
        )

        # This program draws lines from the player/origin
        # to sprites that are within a certain distance.
        # The main action here happens in the geometry shader.
        # It creates lines when a sprite is within the maxDistance.
        self.program_visualize_dist = self.ctx.program(
            vertex_shader="""
            #version 330

            // Sprite positions from SpriteList
            in vec3 in_pos;

            // Output to geometry shader
            out vec3 v_position;

            void main() {
                // This shader just forwards info to geo shader
                v_position = in_pos;
            }
            """,
            geometry_shader="""
            #version 330

            // This is how we access Arcade's global projection matrix
            uniform WindowBlock {
                mat4 projection;
                mat4 view;
            } window;

            // The position we measure distance from
            uniform vec2 origin;
            // The maximum distance
            uniform float maxDistance;

            // These configure the geometry shader to process a points
            // and allows it to emit lines. It runs for every sprite
            // in the spritelist.
            layout (points) in;
            layout (line_strip, max_vertices = 2) out;

            // The position input from vertex shader.
            // It's an array because geo shader can take more than one input
            in vec3 v_position[];

            void main() {
                // ONLY emit a line between the sprite and origin when within the distance
                mat4 mvp = window.projection * window.view;

                if (distance(v_position[0].xy, origin) < maxDistance) {
                    // First line segment position (origin)
                    gl_Position = mvp * vec4(origin, 0.0, 1.0);
                    EmitVertex();
                    // Second line segment position (sprite position)
                    gl_Position = mvp * vec4(v_position[0].xy, 0.0, 1.0);
                    EmitVertex();
                }
            }
            """,
            fragment_shader="""
            #version 330
            // The fragment shader just runs for every pixel of the line segment.

            // Reference to the pixel we are writing to in the framebuffer
            out vec4 fragColor;

            void main() {
                // All the pixels in the line should just be white
                fragColor = vec4(1.0, 1.0, 1.0, 1.0);
            }
            """,
        )
        # Configure program with maximum distance
        self.program_visualize_dist["maxDistance"] = INTERACTION_RADIUS

    def on_draw(self):
        self.clear()
        self.coins.draw()
        # We already have a geometry instance in the spritelist we can
        # use to run our shader/gpu program. It only requires that we
        # use correctly named input name(s). in_pos in this example
        # what will automatically map in the position buffer to the vertex shader.
        self.coins.geometry.render(self.program_visualize_dist, vertices=len(self.coins))
        arcade.draw_sprite(self.player)

        # Visualize the interaction radius
        arcade.draw_circle_filled(
            center_x=self.player.center_x,
            center_y=self.player.center_y,
            radius=INTERACTION_RADIUS,
            color=(255, 255, 255, 64)
        )

    def on_mouse_motion(self, x: float, y: float, dx: float, dy: float):
        # Move the sprite to mouse position
        self.player.position = x, y
        # Update the program with a new origin
        self.program_visualize_dist["origin"] = x, y


SpriteListInteraction().run()



"""
Blending two Backgrounds.

This program loads two infinite scrolling backgrounds.

It then blends between the two when the player reaches a certain x pos.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.background_blending
"""

import arcade
import arcade.future.background as background

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720

WINDOW_TITLE = "Blending Backgrounds Example"

PLAYER_SPEED = 300
CAMERA_SPEED = 0.1


class GameView(arcade.View):
    def __init__(self):
        super().__init__()
        self.camera = arcade.Camera2D()

        # Load the first background from file. Sized to match the screen
        self.background_1 = background.Background.from_file(
            ":resources:/images/tiles/sandCenter.png",
            size=(WINDOW_WIDTH, WINDOW_HEIGHT),
        )

        # Load the second background from file. Sized to match the screen
        self.background_2 = background.Background.from_file(
            ":resources:/images/tiles/dirtCenter.png",
            size=(WINDOW_WIDTH, WINDOW_HEIGHT),
        )

        # start with the second background being hidden
        self.background_2.blend = 0.0

        # Create the player sprite.
        self.player_sprite = arcade.SpriteSolidColor(20, 30, color=arcade.color.PURPLE)
        self.player_sprite.center_y = self.camera.viewport_height // 2
        self.player_sprite.center_x = self.camera.viewport_width // 2

        # Track Player Motion
        self.x_direction = 0
        self.y_direction = 0

    def pan_camera_to_player(self):
        # This will center the camera on the player.
        target_x = self.player_sprite.center_x
        target_y = self.player_sprite.center_y

        # This limits where the player can see. Ensuring they never go too far
        # from the transition.
        if 0.0 > target_x:
            target_x = 0.0
        elif target_x > self.background_1.size[0] * 2:
            target_x = self.background_1.size[0] * 2

        if 0.0 > target_y:
            target_y = 0.0
        elif target_y > self.background_1.size[1]:
            target_y = self.background_1.size[1]

        self.camera.position = arcade.math.lerp_2d(
            self.camera.position,
            (target_x, target_y),
            CAMERA_SPEED,
        )

    def on_update(self, delta_time: float):
        new_position = (
            self.player_sprite.center_x + self.x_direction * delta_time,
            self.player_sprite.center_y + self.y_direction * delta_time,
        )
        self.player_sprite.position = new_position

        # When the player is near x = WINDOW_WIDTH we transition between the two backgrounds.
        if WINDOW_WIDTH - 100 < self.player_sprite.center_x < WINDOW_WIDTH + 100:
            percent = (self.player_sprite.center_x - (WINDOW_WIDTH - 50)) / 200
            self.background_1.blend_layer(self.background_2, percent)

        self.pan_camera_to_player()

    def on_draw(self):
        self.clear()

        with self.camera.activate():
            # Ensure the background aligns with the camera
            self.background_1.pos = self.camera.bottom_left
            self.background_2.pos = self.camera.bottom_left

            # Offset the background texture.
            self.background_1.texture.offset = self.camera.position
            self.background_2.texture.offset = self.camera.position

            self.background_1.draw()
            self.background_2.draw()
            arcade.draw_sprite(self.player_sprite)

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol in (arcade.key.LEFT, arcade.key.A):
            self.x_direction -= PLAYER_SPEED
        elif symbol in (arcade.key.RIGHT, arcade.key.D):
            self.x_direction += PLAYER_SPEED
        elif symbol in (arcade.key.DOWN, arcade.key.S):
            self.y_direction -= PLAYER_SPEED
        elif symbol in (arcade.key.UP, arcade.key.W):
            self.y_direction += PLAYER_SPEED
        # Close the window
        elif symbol == arcade.key.ESCAPE:
            self.window.close()

    def on_key_release(self, symbol: int, modifiers: int):
        if symbol in (arcade.key.LEFT, arcade.key.A):
            self.x_direction += PLAYER_SPEED
        elif symbol in (arcade.key.RIGHT, arcade.key.D):
            self.x_direction -= PLAYER_SPEED
        elif symbol in (arcade.key.DOWN, arcade.key.S):
            self.y_direction += PLAYER_SPEED
        elif symbol in (arcade.key.UP, arcade.key.W):
            self.y_direction -= PLAYER_SPEED

    def on_resize(self, width: int, height: int):
        super().on_resize(width, height)
        self.camera.match_window()

        # This is to ensure the background covers the entire screen.
        self.background_1.size = (width, height)
        self.background_2.size = (width, height)


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE, resizable=True)

    # Create and setup the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Minimal Sprite Example

Draws a single sprite in the middle screen.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_minimal
"""
import arcade


class GameView(arcade.View):

    def __init__(self):
        super().__init__()
        # 1. Create the SpriteList
        self.sprites = arcade.SpriteList()

        # 2. Create & append your Sprite instance to the SpriteList
        self.circle = arcade.Sprite()  # Sprite with the default texture
        self.circle.position = self.center  # center sprite on screen
        self.sprites.append(self.circle)  # Append the instance to the SpriteList

    def on_draw(self):
        # 3. Clear the screen
        self.clear()

        # 4. Call draw() on the SpriteList inside an on_draw() method
        self.sprites.draw()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(1280, 720, "Minimal SPrite Example")

    # Create and setup the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from arcade import gl
from arcade.earclip import earclip
from arcade.types import Point2, Point2List, RGBOrA255

from .helpers import _generic_draw_line_strip, get_points_for_thick_line


def draw_polygon_filled(point_list: Point2List, color: RGBOrA255) -> None:
    """
    Draw a polygon that is filled in.

    Args:
        point_list:
            List of points making up the lines. Each point is
            in a list. So it is a list of lists.
        color:
            The color, specified in RGB or RGBA format.
    """
    triangle_points = earclip(point_list)
    flattened_list = tuple(i for g in triangle_points for i in g)
    _generic_draw_line_strip(flattened_list, color, gl.TRIANGLES)


def draw_polygon_outline(point_list: Point2List, color: RGBOrA255, line_width: float = 1.0) -> None:
    """
    Draw a polygon outline. Also known as a "line loop."

    Args:
        point_list:
            List of points making up the lines. Each point is
            in a list. So it is a list of lists.
        color:
            The color of the outline as an RGBA :py:class:`tuple` or
            :py:class:`.Color` instance.
        line_width:
            Width of the line in pixels.
    """
    # Convert to modifiable list & close the loop
    new_point_list = list(point_list)
    new_point_list.append(point_list[0])

    # Create a place to store the triangles we'll use to thicken the line
    triangle_point_list: list[Point2] = []

    # This needs a lot of improvement
    last_point = None
    for point in new_point_list:
        if last_point is not None:
            # Calculate triangles, then re-order to link up the quad?
            points = get_points_for_thick_line(*last_point, *point, line_width)
            reordered_points = points[1], points[0], points[2], points[3]

            triangle_point_list.extend(reordered_points)
        last_point = point

    # Use first two points of new list to close the loop
    new_start, new_next = new_point_list[:2]
    s_x, s_y = new_start
    n_x, n_y = new_next
    points = get_points_for_thick_line(s_x, s_y, n_x, n_y, line_width)
    triangle_point_list.append(points[1])

    _generic_draw_line_strip(triangle_point_list, color, gl.TRIANGLE_STRIP)



import arcade.gl as gl
from arcade.future.background import BackgroundTexture
from arcade.window_commands import get_window


class Background:
    """
    Backgrounds are large geometries to which a Background texture is rendered.
    By default, the position defines the bottom left corner.
    If the size is larger than the given BackgroundTexture the texture will repeat.
    A shift value can be given when calling draw.
    This can be used to move the background without actually adjusting the position
    You may supply your own shader and geometries.
    The default shader implements 4 uniforms.
    vec2 pos, vec2 size, vec3 color, mat3 pixelTransform, and float blend.
    """

    def __init__(
        self,
        texture: BackgroundTexture,
        pos: tuple[float, float],
        size: tuple[int, int],
        color: tuple[float, float, float] | tuple[int, int, int],
        shader: gl.Program | None = None,
        geometry: gl.Geometry | None = None,
    ):
        self._ctx = get_window().ctx
        if shader is None:
            shader = self._ctx.load_program(
                vertex_shader=":system:/shaders/background_vs.glsl",
                fragment_shader=":system:/shaders/background_fs.glsl",
            )
        self.shader = shader

        if geometry is None:
            geometry = gl.geometry.quad_2d(pos=(0.5, 0.5))
        self.geometry = geometry

        self.texture = texture

        self._pos = pos
        try:
            self.shader["pos"] = pos
        except KeyError:
            print(
                "Attempting to set uniform 'pos' when the shader does not "
                "have a uniform with that name."
            )

        self._size = size
        try:
            self.shader["size"] = size
        except KeyError:
            print(
                "Attempting to set uniform 'size' when the shader does not "
                "have a uniform with that name."
            )

        self._blend = 1.0
        try:
            self.shader["blend"] = 1.0
        except KeyError:
            print(
                "Attempting to set uniform 'blend' when the shader does not "
                "have a uniform with that name."
            )

        self._color = (
            color if sum(color) <= 3.0 else (color[0] / 255, color[1] / 255, color[2] / 255)
        )
        try:
            self.shader["color"] = self._color
        except KeyError:
            print(
                "Attempting to set uniform 'color' when the shader does not "
                "have a uniform with that name."
            )

    @staticmethod
    def from_file(
        tex_src: str,
        pos: tuple[float, float] = (0.0, 0.0),
        size: tuple[int, int] | None = None,
        offset: tuple[float, float] = (0.0, 0.0),
        scale: float = 1.0,
        angle: float = 0.0,
        *,
        filters=(gl.NEAREST, gl.NEAREST),
        color: tuple[int, int, int] | None = None,
        color_norm: tuple[float, float, float] | None = None,
        shader: gl.Program | None = None,
        geometry: gl.Geometry | None = None,
    ):
        """
        This will generate a Background from an input image source.
        The generated texture is not stored in the texture cache or any texture atlas.

        Args:
            tex_src: The image source.
            pos: The position of the Background (Bottom Left Corner by default).
            size: The width and height of the Background.
            offset: The BackgroundTexture offset.
            scale: The BackgroundTexture Scale.
            angle: The BackgroundTexture angle.
            filters: The OpenGl Texture filters (gl.Nearest by default).
            color: This is a color defined from 0-255. Priorities color_norm
            color_norm: This is a color defined from 0.0-1.0. Priorities color_norm
                            assumed to be in the range 0.0-1.0.
            shader: The shader used for rendering.
            geometry: The geometry used for rendering (a rectangle equal to the size by default).
        """
        background_texture = BackgroundTexture.from_file(tex_src, offset, scale, angle, filters)
        if size is None:
            size = background_texture.texture.size

        if color_norm:
            _color = color_norm
        elif color:
            _color = color[0] / 255, color[1] / 255, color[2] / 255
        else:
            _color = (1.0, 1.0, 1.0)

        return Background(background_texture, pos, size, _color, shader, geometry)

    @property
    def pos(self) -> tuple[float, float]:
        return self._pos

    @pos.setter
    def pos(self, value: tuple[float, float]):
        self._pos = value

    @property
    def size(self) -> tuple[int, int]:
        return self._size

    @size.setter
    def size(self, value: tuple[int, int]):
        self._size = value
        try:
            self.shader["size"] = value
        except KeyError:
            print(
                "Attempting to set uniform 'size' when the shader does not "
                "have a uniform with that name."
            )

    @property
    def blend(self) -> float:
        return self._blend

    @blend.setter
    def blend(self, value):
        self._blend = value
        try:
            self.shader["blend"] = value
        except KeyError:
            print(
                "Attempting to set uniform 'blend' when the shader does not "
                "have a uniform with that name."
            )

    @property
    def color(self) -> tuple[int, int, int]:
        """
        Color in the range of 0-255.
        """
        return (
            int(self._color[0] * 255),
            int(self._color[1] * 255),
            int(self._color[2] * 255),
        )

    @color.setter
    def color(self, value: tuple[int, int, int]):
        """
        Color in the range of 0-255.
        """
        self._color = (value[0] / 255, value[1] / 255, value[2] / 255)
        try:
            self.shader["color"] = self._color
        except KeyError:
            print(
                "Attempting to set uniform 'color' when shader does not "
                "have uniform with that name."
            )

    @property
    def color_norm(self) -> tuple[float, float, float]:
        return self._color

    @color_norm.setter
    def color_norm(self, value: tuple[float, float, float]):
        self._color = value
        try:
            self.shader["color"] = self._color
        except KeyError:
            print(
                "Attempting to set uniform 'color' when shader does not "
                "have uniform with that name."
            )

    def draw(self, shift: tuple[float, float] = (0.0, 0.0)):
        try:
            self.shader["pixelTransform"] = self.texture.pixel_transform
        except KeyError:
            print(
                "Attempting to set uniform 'pixelTransform' when the shader does "
                "not have a uniform with that name."
            )

        try:
            self.shader["pos"] = self.pos[0] + shift[0], self.pos[1] + shift[1]
        except KeyError:
            print(
                "Attempting to set uniform 'pos' when the shader does not have a "
                "uniform with that name."
            )

        self.texture.use(0)
        with self._ctx.enabled(self._ctx.BLEND):
            self.geometry.render(self.shader)

    def blend_layer(self, other, percent: float):
        self.blend = 1 - percent
        other.blend = percent



"""
Pymunk Physics Engine
"""

import logging
import math
from collections.abc import Callable

import pymunk
from pyglet.math import Vec2

from arcade import Sprite

__all__ = ["PymunkPhysicsObject", "PymunkException", "PymunkPhysicsEngine"]

from arcade.utils import copy_dunders_unimplemented

LOG = logging.getLogger(__name__)


class PymunkPhysicsObject:
    """Object that holds pymunk body/shape for a sprite."""

    def __init__(self, body: pymunk.Body | None = None, shape: pymunk.Shape | None = None):
        self.body: pymunk.Body | None = body
        self.shape: pymunk.Shape | None = shape


class PymunkException(Exception):
    """Exception raised for errors in the PymunkPhysicsEngine."""

    pass


# Temp fix for https://github.com/pythonarcade/arcade/issues/2074
@copy_dunders_unimplemented
class PymunkPhysicsEngine:
    """An Arcade-specific adapter for Pymunk.

    `Pymunk`_ is itself a Python adapter for the professional-grade
    `Chipmunk2D`_ engine. However, Arcade's ``PymunkPhysicsEngine``
    and its doc are currently in need of improvement.

     .. note:: Arcade would welcome assistance with improving it.

               If you are interested, please see Arcade's
               `CONTRIBUTING.md <CONTRIBUTING.md: https://github.com/pythonarcade/arcade/blob/development/CONTRIBUTING.md>`_

    Args:
        gravity:
            The direction where gravity is pointing.
            See :py:attr:`pymunk.Space.gravity` to learn more.
        damping:
            The default velocity loss per tick across the
            :py:class:`~pymunk.Space` for all :py:attr:`DYNAMIC`
            objects.

            * Override this for objects by passing different value
              :`add_sprite` or :py:meth:`add_spritelist`
            * See :py:attr:`pymunk.Space.damping` to learn more

        maximum_incline_on_ground:
            The maximum incline the ground can have before
            :py:meth:`is_on_ground` returns ``False``.

            * Defaults to ``0.708`` radians (a bit over 45 °)
            * Not a pymunk value, but an Arcade feature

    """

    DYNAMIC = pymunk.Body.DYNAMIC
    """A ``body_type`` for moving Pymunk-controlled objects.

    An indirect approach is best for controlling the velocity and
    positioning of dynamic objects:

    * :py:meth:`apply_force`
    * :py:meth:`apply_impulse`

    .. warning:: Avoid setting velocity directly on dynamic objects!

                 If you need to set velocity directly, you may want to
                 pass :py:attr:`KINEMATIC` as the ``body_type`` to
                 :py:meth:`add_sprite` instead.

    If you :py:class:`set_velocity` directly anyway, the
    following may occur:

    #. Setting velocity approaches infinite acceleration
    #. ``f = m * a`` approaches ``f = m * infinity``
    #. Collisions go haywire

    In some games, you may be able to find a way to harness this for
    comedic effect.

    .. note:: This value is an alias of :py:attr:`pymunk.Body.DYNAMIC`.

              Please see the Pymunk page linked above to learn more.
    """
    STATIC = pymunk.Body.STATIC
    """A ``body_type`` for objects which do not move.

    This is best used for terrain or non-moving platforms.

    .. note:: This value is an alias of :py:attr:`pymunk.Body.STATIC`.

              Please see the Pymunk page linked above to learn more.
    """
    KINEMATIC = pymunk.Body.KINEMATIC
    """A ``body_type`` for objects controlled by your code or Arcade's.

    When colliding, Kinematic objects:

    * act as if they have infinite mass
    * prevent joined and touching objects from sleeping

    This makes them excellent for game elements like moving platforms or
    hazards which move or crush game objects. You can control kinematic
    objects by setting their positions and velocities directly:

    * :py:meth:`set_velocity`
    * :py:meth:`set_velocity_horizontal`
    * :py:meth:`set_velocity_vertical`
    * :py:meth:`set_position`


    .. note:: This value is an alias of :py:attr:`pymunk.Body.KINEMATIC`.

              Please see the Pymunk page linked above to learn more.
    """
    MOMENT_INF = float("inf")

    def __init__(
        self, gravity=(0, 0), damping: float = 1.0, maximum_incline_on_ground: float = 0.708
    ) -> None:
        # -- Pymunk
        self.space = pymunk.Space()
        self.space.gravity = gravity
        self.space.damping = damping
        self.collision_types: list[str] = []
        self.sprites: dict[Sprite, PymunkPhysicsObject] = {}
        self.non_static_sprite_list: list = []
        self.maximum_incline_on_ground = maximum_incline_on_ground

    def add_sprite(
        self,
        sprite: Sprite,
        mass: float = 1.0,
        friction: float = 0.2,
        elasticity: float | None = None,
        moment_of_inertia: float | None = None,  # correct spelling
        body_type: int = DYNAMIC,
        damping: float | None = None,
        gravity: pymunk.Vec2d | tuple[float, float] | Vec2 | None = None,
        max_velocity: int | None = None,
        max_horizontal_velocity: int | None = None,
        max_vertical_velocity: int | None = None,
        radius: float = 0,
        collision_type: str | None = "default",
    ) -> None:
        """Add a sprite to the physics engine.

        Args:
            sprite:
                A :py:class:`.Sprite` to add
            mass:
                The mass of the object (Defaults to ``1.0``).
            friction:
                How much the object resists sliding against surfaces:

                .. list-table::
                   :header-rows: 0

                   * - ``0.0``
                     - Absolutely slippery with no resistance at all
                   * - ``0.2``
                     - Default (Waxed wood on very wet snow)
                   * - ``friction > 1.0``
                     - Very rough

                *Higher values may not make a meaningful difference.*

                See :py:attr:`pymunk.Shape.friction` to learn more.

            elasticity:
                How bouncy the object is.

                .. list-table::
                   :header-rows: 0

                   * - ``0.0``
                     - No bounce
                   * - ``0.99``
                     - Very bouncy
                   * - ``elasticity >= 1.0``
                     - May behave badly (breaks conservation of energy)

                See :py:attr:`pymunk.Shape.elasticity` to learn more.

            moment_of_inertia:
                How much force is needed to change the object's rotation (
                pass :py:attr:`MOMENT_INF` or ``float('inf')`` to "lock"
                its angle).

                See :py:attr:`pymunk.Shape.moment_of_inertia` to learn more.

            body_type:
                :py:attr:`DYNAMIC` (default), :py:attr:`KINEMATIC`, or
                :py:attr:`STATIC`.
            damping:
                Like air resistance. See the :py:class:`.PymunkPhysicsEngine`
                top-level doc.
            gravity:
                See the :py:class:`.PymunkPhysicsEngine` top-level doc.
            max_velocity:
                The maximum velocity of this object.
            max_horizontal_velocity:
                Clamp the velocity on the x axis to this.
            max_vertical_velocity:
                Clamp the velocity along the y axis to this.
            radius:
                The radius for the :py:class:`pymunk.Shape` created for
                the :py:class:`sprite <.Sprite>`.
            collision_type:
                Assign a collision name to this sprite. It will be used
                by :py:meth:`add_collision_handler` if called.
        """

        if damping is not None:
            sprite.pymunk.damping = damping  # pyright: ignore [reportGeneralTypeIssues=false]

        if gravity is not None:
            sprite.pymunk.gravity = gravity  # pyright: ignore [reportGeneralTypeIssues=false]

        if max_velocity is not None:
            sprite.pymunk.max_velocity = (  # pyright: ignore [reportGeneralTypeIssues=false]
                max_velocity
            )

        if max_vertical_velocity is not None:
            sprite.pymunk.max_vertical_velocity = max_vertical_velocity  # pyright: ignore

        if max_horizontal_velocity is not None:
            sprite.pymunk.max_horizontal_velocity = max_horizontal_velocity  # pyright: ignore

        # See if the sprite already has been added
        if sprite in self.sprites:
            LOG.warning("Attempt to add a Sprite that has already been added. Ignoring.")
            return

        # Keep track of collision types
        if collision_type not in self.collision_types:
            # LOG.debug(f"Adding new collision type of {collision_type}.")
            self.collision_types.append(collision_type)  # type: ignore

        # Get a number associated with the string of collision_type
        collision_type_id = self.collision_types.index(collision_type)  # type: ignore

        # Default to a box moment_of_inertia
        if moment_of_inertia is None:
            moment_of_inertia = pymunk.moment_for_box(mass, (sprite.width, sprite.height))

        # Create the physics body
        body = pymunk.Body(mass, moment_of_inertia, body_type=body_type)

        # Set the body's position
        body.position = pymunk.Vec2d(sprite.center_x, sprite.center_y)
        body.angle = math.radians(sprite.angle)

        # Callback used if we need custom gravity, damping, velocity, etc.
        def velocity_callback(
            my_body: pymunk.Body, my_gravity: tuple[float, float], my_damping: float, dt: float
        ):
            """Used for custom damping, gravity, and max_velocity."""

            # Custom damping
            if sprite.pymunk.damping is not None:
                adj_damping = ((sprite.pymunk.damping * 100.0) / 100.0) ** dt
                # print(
                #     f"Custom damping {sprite.pymunk.damping} {my_damping} "
                #     "default to {adj_damping}"
                # )
                my_damping = adj_damping

            # Custom gravity
            if sprite.pymunk.gravity is not None:
                my_gravity = sprite.pymunk.gravity

            # Go ahead and update velocity
            pymunk.Body.update_velocity(my_body, my_gravity, my_damping, dt)

            # Now see if we are going too fast...

            # Support max velocity
            if sprite.pymunk.max_velocity:
                velocity = my_body.velocity.length
                if velocity > sprite.pymunk.max_velocity:
                    scale = sprite.pymunk.max_velocity / velocity
                    my_body.velocity = my_body.velocity * scale

            # Support max horizontal velocity
            if sprite.pymunk.max_horizontal_velocity:
                velocity = my_body.velocity.x
                if abs(velocity) > sprite.pymunk.max_horizontal_velocity:
                    velocity = sprite.pymunk.max_horizontal_velocity * math.copysign(1, velocity)
                    my_body.velocity = pymunk.Vec2d(velocity, my_body.velocity.y)

            # Support max vertical velocity
            if max_vertical_velocity:
                velocity = my_body.velocity[1]
                if abs(velocity) > max_vertical_velocity:
                    velocity = max_vertical_velocity * math.copysign(1, velocity)
                    my_body.velocity = pymunk.Vec2d(my_body.velocity.x, velocity)

        # Add callback if we need to do anything custom on this body
        # if damping or gravity or max_velocity or max_horizontal_velocity or max_vertical_velocity:
        if body_type == self.DYNAMIC:
            body.velocity_func = velocity_callback

        # Set the physics shape to the sprite's hitbox
        poly = sprite.hit_box.points
        scaled_poly = [[x * sprite.scale_x for x in z] for z in poly]
        shape = pymunk.Poly(body, scaled_poly, radius=radius)  # type: ignore

        # Set collision type, used in collision callbacks
        if collision_type:
            shape.collision_type = collision_type_id

        # How bouncy is the shape?
        if elasticity is not None:
            shape.elasticity = elasticity

        # Set shapes friction
        shape.friction = friction

        # Create physics object and add to list
        physics_object = PymunkPhysicsObject(body, shape)
        self.sprites[sprite] = physics_object
        if body_type != self.STATIC:
            self.non_static_sprite_list.append(sprite)

        # Add body and shape to pymunk engine
        self.space.add(body, shape)

        # Register physics engine with sprite, so we can remove from physics engine
        # if we tell the sprite to go away.
        sprite.register_physics_engine(self)

    def add_sprite_list(
        self,
        sprite_list,
        mass: float = 1,
        friction: float = 0.2,
        elasticity: float | None = None,
        moment_of_inertia: float | None = None,
        body_type: int = DYNAMIC,
        damping: float | None = None,
        collision_type: str | None = None,
    ) -> None:
        """
        Add all sprites in a sprite list to the physics engine.

        Args:
            sprite_list:
                A list of sprites to add
            mass:
                The mass of the object (Defaults to ``1.0``).
            friction:
                How much the object resists sliding against surfaces:

                .. list-table::
                   :header-rows: 0

                   * - ``0.0``
                     - Absolutely slippery with no resistance at all
                   * - ``0.2``
                     - Default (Waxed wood on very wet snow)
                   * - ``friction > 1.0``
                     - Very rough

                *Higher values may not make a meaningful difference.*

                See :py:attr:`pymunk.Shape.friction` to learn more.

            elasticity:
                How bouncy the object is.

                .. list-table::
                   :header-rows: 0

                   * - ``0.0``
                     - No bounce
                   * - ``0.99``
                     - Very bouncy
                   * - ``elasticity >= 1.0``
                     - May behave badly (breaks conservation of energy)

                See :py:attr:`pymunk.Shape.elasticity` to learn more.

            moment_of_inertia:
                How much force is needed to change the object's rotation (
                pass :py:attr:`MOMENT_INF` or ``float('inf')`` to "lock"
                its angle).

                See :py:attr:`pymunk.Shape.moment_of_inertia` to learn more.

            body_type:
                :py:attr:`DYNAMIC` (default), :py:attr:`KINEMATIC`, or
                :py:attr:`STATIC`.
            damping:
                Like air resistance. See the :py:class:`.PymunkPhysicsEngine`
                top-level doc.
            collision_type:
                Assign a collision name to this sprite. It will be used
                by :py:meth:`add_collision_handler` if called.
        """
        for sprite in sprite_list:
            self.add_sprite(
                sprite=sprite,
                mass=mass,
                friction=friction,
                elasticity=elasticity,
                moment_of_inertia=moment_of_inertia,
                body_type=body_type,
                damping=damping,
                collision_type=collision_type,
            )

    def remove_sprite(self, sprite: Sprite) -> None:
        """Remove a sprite from the physics engine."""
        physics_object = self.sprites[sprite]
        self.space.remove(physics_object.body)  # type: ignore
        self.space.remove(physics_object.shape)  # type: ignore
        self.sprites.pop(sprite)
        if sprite in self.non_static_sprite_list:
            self.non_static_sprite_list.remove(sprite)

    def get_sprite_for_shape(self, shape: pymunk.Shape | None) -> Sprite | None:
        """Try to get the sprite registered with this engine for ``shape``.

        This method returns ``None`` when:

        * ``shape`` is ``None``
        * No :py:class:`.Sprite` was to this engine for ``shape``

        The second item may occur if you are using multiple instances of
        :py:class:`.PymunkPhysicsEngine`.

        Args:
            shape:
                A Pymunk shape to perform lookup for.
        Returns:
            A sprite for the ``shape``; ``None`` if no sprite is known.
        """
        for sprite in self.sprites:
            if self.sprites[sprite].shape is shape:
                return sprite
        return None

    def get_sprites_from_arbiter(
        self, arbiter: pymunk.Arbiter
    ) -> tuple[Sprite | None, Sprite | None]:
        """Given a collision arbiter, return the sprites associated with the collision."""
        shape1, shape2 = arbiter.shapes
        sprite1 = self.get_sprite_for_shape(shape1)
        sprite2 = self.get_sprite_for_shape(shape2)
        return sprite1, sprite2

    def is_on_ground(self, sprite: Sprite) -> bool:
        """Return true of sprite is on top of something."""
        grounding = self.check_grounding(sprite)
        return grounding["body"] is not None

    def apply_impulse(self, sprite: Sprite, impulse: tuple[float, float]) -> None:
        """Apply an impulse force on a sprite"""
        physics_object = self.get_physics_object(sprite)
        if physics_object.body is None:
            raise PymunkException(
                "Tried to apply an impulse, but this physics object has no 'body' set."
            )
        physics_object.body.apply_impulse_at_local_point(impulse)

    def set_position(self, sprite: Sprite, position: pymunk.Vec2d | tuple[float, float]):
        """Set the position of the sprite in the engine's simulation.

        To learn more, please see :py:attr:`pymunk.Body.position`.

        Args:
            sprite:
                An Arcade :py:class:`.Sprite` known to the engine.
            position:
                A two-dimensional position in world space.
        """
        physics_object = self.get_physics_object(sprite)
        if physics_object.body is None:
            raise PymunkException(
                "Tried to set a position, but this physics object has no 'body' set."
            )
        physics_object.body.position = position

    def set_rotation(self, sprite: Sprite, rotation: float) -> None:
        """
        Set the rotation of the sprite

        Args:
            sprite:
                A sprite known to the physics engine.
            rotation:
                The angle in degrees (clockwise).
        """
        physics_object = self.get_physics_object(sprite)
        if physics_object.body is None:
            raise PymunkException(
                "Tried to set a rotation, but this physics object has no 'body' set."
            )
        physics_object.body.angle = math.radians(rotation)

    def set_velocity(self, sprite: Sprite, velocity: tuple[float, float]) -> None:
        """Directly set the velocity of a sprite known to the engine.

        .. warning:: Avoid using this on any :py:attr:`DYNAMIC` objects!

        This function is meant for :py:attr:`KINEMATIC` objects. Using
        it on a sprite added as :py:attr:`DYNAMIC` can cause strange and
        very broken behavior.

        To learn more, please see:

        * Pymunk's documentation on :py:attr:`~pymunk.Body.DYNAMIC` and
          :py:attr:`~pymunk.Body.KINEMATIC`

        """
        physics_object = self.get_physics_object(sprite)
        if physics_object.body is None:
            raise PymunkException(
                "Tried to set a velocity, but this physics object has no 'body' set."
            )
        physics_object.body.velocity = velocity

    def add_collision_handler(
        self,
        first_type: str,
        second_type: str,
        begin_handler: Callable | None = None,
        pre_handler: Callable | None = None,
        post_handler: Callable | None = None,
        separate_handler: Callable | None = None,
    ) -> None:
        """
        Add code to handle collisions between objects.

        Args:
            first_type: The first type of object to check for collisions.
            second_type: The second type of object to check for collisions.
            begin_handler: Function to call when a collision begins.
            pre_handler: Function to call before a collision is resolved.
            post_handler: Function to call after a collision is resolved.
            separate_handler: Function to call when two objects
        """

        if first_type not in self.collision_types:
            # LOG.debug(f"Adding new collision type of {first_type}.")
            self.collision_types.append(first_type)
        first_type_id = self.collision_types.index(first_type)

        if second_type not in self.collision_types:
            # LOG.debug(f"Adding new collision type of {second_type}.")
            self.collision_types.append(second_type)
        second_type_id = self.collision_types.index(second_type)

        def _f1(arbiter, space, data):
            sprite_a, sprite_b = self.get_sprites_from_arbiter(arbiter)
            should_process_collision = False
            if sprite_a is not None and sprite_b is not None and begin_handler is not None:
                should_process_collision = begin_handler(sprite_a, sprite_b, arbiter, space, data)
            return should_process_collision

        def _f2(arbiter, space, data):
            sprite_a, sprite_b = self.get_sprites_from_arbiter(arbiter)
            if sprite_a is not None and sprite_b is not None and post_handler is not None:
                post_handler(sprite_a, sprite_b, arbiter, space, data)

        def _f3(arbiter, space, data):
            sprite_a, sprite_b = self.get_sprites_from_arbiter(arbiter)
            if pre_handler is not None:
                return pre_handler(sprite_a, sprite_b, arbiter, space, data)

        def _f4(arbiter, space, data):
            sprite_a, sprite_b = self.get_sprites_from_arbiter(arbiter)
            if separate_handler:
                separate_handler(sprite_a, sprite_b, arbiter, space, data)

        h = self.space.add_collision_handler(first_type_id, second_type_id)
        if begin_handler:
            h.begin = _f1
        if post_handler:
            h.post_solve = _f2
        if pre_handler:
            h.pre_solve = _f3
        if separate_handler:
            h.separate = _f4

    def resync_sprites(self) -> None:
        """
        Set visual sprites to be the same location as physics engine sprites.
        Call this after stepping the pymunk physics engine
        """
        # Create copy in case a sprite wants to remove itself from the list as
        # we iterate through the list.
        sprites = self.non_static_sprite_list.copy()

        for sprite in sprites:
            # Get physics object for this sprite
            physics_object = self.sprites[sprite]

            original_position = sprite.position

            if physics_object.body:
                # Item is sleeping, skip
                if physics_object.body.is_sleeping:
                    continue

                new_position = physics_object.body.position
                new_angle = -math.degrees(physics_object.body.angle)

                # Calculate change in location, used in call-back
                dx = new_position[0] - original_position[0]
                dy = new_position[1] - original_position[1]
                d_angle = new_angle - sprite.angle

                # Update sprite to new location
                sprite.position = new_position
                sprite.angle = new_angle

                # Notify sprite we moved, in case animation needs to be updated
                sprite.pymunk_moved(self, dx, dy, d_angle)

    def step(self, delta_time: float = 1 / 60.0, resync_sprites: bool = True) -> None:
        """
        Tell the physics engine to perform calculations.

        Args:
            delta_time: Time to move the simulation forward. Keep this
                value constant, do not use varying values for each step.
            resync_sprites: Resynchronize Arcade graphical sprites to be
                at the same location as their Pymunk counterparts.
                If running multiple steps per frame, set this to
                false for the first steps, and true for the last
                step that's part of the update.
        """
        # Update physics
        # Use a constant time step, don't use delta_time
        # See "Game loop / moving time forward"
        # http://www.pymunk.org/en/latest/overview.html#game-loop-moving-time-forward
        self.space.step(delta_time)
        if resync_sprites:
            self.resync_sprites()

    def get_physics_object(self, sprite: Sprite) -> PymunkPhysicsObject:
        """
        Get the shape/body for a sprite.

        Args:
            sprite:
                The sprite to get the physics object for.
        """
        return self.sprites[sprite]

    def apply_force(self, sprite: Sprite, force: tuple[float, float]):
        """
        Apply force to a Sprite.

        Args:
            sprite:
                The sprite to apply the force to.
            force:
                The force to apply to the sprite.
        """
        physics_object = self.sprites[sprite]
        if physics_object.body is None:
            raise PymunkException(
                "Tried to apply a force, but this physics object has no 'body' set."
            )
        physics_object.body.apply_force_at_local_point(force, (0, 0))

    def set_horizontal_velocity(self, sprite: Sprite, velocity: float) -> None:
        """
        Set a sprite's velocity.

        Args:
            sprite:
                The sprite to set the velocity for.
            velocity:
                The velocity to set the sprite to.
        """
        physics_object = self.sprites[sprite]
        if physics_object.body is None:
            raise PymunkException(
                "Tried to set a velocity, but this physics object has no 'body' set."
            )
        cv = physics_object.body.velocity
        new_cv = (velocity, cv[1])
        physics_object.body.velocity = new_cv

    def set_friction(self, sprite: Sprite, friction: float) -> None:
        """Set the friction a sprite experiences against other surfaces.

        This is how "rough" a sprite is during a collision with others:

        * ``0.0`` is the lowest value allowed (absolute slipperiness)
        * Higher values slide less on surfaces and other objects

        Pymunk allows setting ``friction`` higher than ``1.0``, but very
        high values might not have meaningful gameplay impact.

        .. _Simple Wikipedia's Article on Friction: https://simple.wikipedia.org/wiki/Friction

        To learn more, please see:

        * The :ref:`pymunk_platformer_tutorial-add_physics_engine` step
          of the :ref:`pymunk_platformer_tutorial`
        * `Simple Wikipedia's Article on Friction`_
        * :py:attr:`pymunk.Poly.friction`

        Args:
            sprite:
                The sprite to set the friction for.
            friction:
                How much the object resists sliding against surfaces.
        """
        physics_object = self.sprites[sprite]
        if physics_object.shape is None:
            raise PymunkException(
                "Tried to set friction, but this physics object has no 'shape' set."
            )
        physics_object.shape.friction = friction

    def apply_opposite_running_force(self, sprite: Sprite) -> None:
        """
        If a sprite goes left while on top of a dynamic sprite, that sprite
        should get pushed to the right.
        """
        grounding = self.check_grounding(sprite)
        body = self.get_physics_object(sprite).body
        if not body:
            raise ValueError("Physics body not set.")

        if body.force[0] and grounding and grounding["body"]:
            grounding["body"].apply_force_at_world_point((-body.force[0], 0), grounding["position"])

    def check_grounding(self, sprite: Sprite) -> dict:
        """
        See if the player is on the ground. Used to see if we can jump.

        Args:
            sprite: The sprite to check if it is on the ground.
        """
        grounding = {
            "normal": pymunk.Vec2d.zero(),
            "penetration": pymunk.Vec2d.zero(),
            "impulse": pymunk.Vec2d.zero(),
            "position": pymunk.Vec2d.zero(),
            "body": None,
        }

        # creates a unit vector (Vector of length 1) in the same direction as the gravity
        gravity_unit_vector = pymunk.Vec2d(1, 0).rotated(self.space.gravity.angle)

        def f(arbiter: pymunk.Arbiter):
            """
            Checks if the the point of collision is in a way, that the sprite is on top of the other
            """
            # Gets the normal vector of the collision. This is the point of collision.
            n = arbiter.contact_point_set.normal

            # Checks if the x component of the gravity is in range of the maximum incline,
            # same for the y component.
            # This will work, as the normal AND gravity are both points on a circle with
            # a radius of 1. (both are unit vectors)
            if (
                gravity_unit_vector.x + self.maximum_incline_on_ground
                > n.x
                > gravity_unit_vector.x - self.maximum_incline_on_ground
                and gravity_unit_vector.y + self.maximum_incline_on_ground
                > n.y
                > gravity_unit_vector.y - self.maximum_incline_on_ground
            ):
                grounding["normal"] = n
                grounding["penetration"] = -arbiter.contact_point_set.points[0].distance
                grounding["body"] = arbiter.shapes[1].body
                grounding["impulse"] = arbiter.total_impulse
                grounding["position"] = arbiter.contact_point_set.points[0].point_b

        physics_object = self.sprites[sprite]
        if not physics_object.body:
            raise ValueError("No physics body set.")
        physics_object.body.each_arbiter(f)

        return grounding



"""
Sprite Bullets

Simple program to show basic sprite usage.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_bullets
"""
import random
import arcade

SPRITE_SCALING_PLAYER = 0.6
SPRITE_SCALING_COIN = 0.4
SPRITE_SCALING_LASER = 0.8
COIN_COUNT = 50

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprites and Bullets Example"

BULLET_SPEED = 5


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """ Initializer """
        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.player_list = None
        self.coin_list = None
        self.bullet_list = None

        # Set up the player info
        self.player_sprite = None
        self.score = 0

        # Don't show the mouse cursor
        self.window.set_mouse_visible(False)

        # Load sounds. Sounds from kenney.nl
        self.gun_sound = arcade.load_sound(":resources:sounds/hurt5.wav")
        self.hit_sound = arcade.load_sound(":resources:sounds/hit5.wav")

        self.background_color = arcade.color.AMAZON

    def setup(self):

        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.player_list = arcade.SpriteList()
        self.coin_list = arcade.SpriteList()
        self.bullet_list = arcade.SpriteList()

        # Set up the player
        self.score = 0

        # Image from kenney.nl
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING_PLAYER)
        self.player_sprite.center_x = 50
        self.player_sprite.center_y = 70
        self.player_list.append(self.player_sprite)

        # Create the coins
        for i in range(COIN_COUNT):

            # Create the coin instance
            # Coin image from kenney.nl
            coin = arcade.Sprite(
                ":resources:images/items/coinGold.png",
                scale=SPRITE_SCALING_COIN,
            )

            # Position the coin
            coin.center_x = random.randrange(WINDOW_WIDTH)
            coin.center_y = random.randrange(120, WINDOW_HEIGHT)

            # Add the coin to the lists
            self.coin_list.append(coin)

        # Set the background color
        self.background_color = arcade.color.AMAZON

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites.
        self.coin_list.draw()
        self.bullet_list.draw()
        self.player_list.draw()

        # Render the text
        arcade.draw_text(f"Score: {self.score}", 10, 20, arcade.color.WHITE, 14)

    def on_mouse_motion(self, x, y, dx, dy):
        """
        Called whenever the mouse moves.
        """
        self.player_sprite.center_x = x

    def on_mouse_press(self, x, y, button, modifiers):
        """
        Called whenever the mouse button is clicked.
        """
        # Gunshot sound
        arcade.play_sound(self.gun_sound)
        # Create a bullet
        bullet = arcade.Sprite(
            ":resources:images/space_shooter/laserBlue01.png",
            scale=SPRITE_SCALING_LASER,
        )

        # The image points to the right, and we want it to point up. So
        # rotate it.
        bullet.angle = -90

        # Give the bullet a speed
        bullet.change_y = BULLET_SPEED

        # Position the bullet
        bullet.center_x = self.player_sprite.center_x
        bullet.bottom = self.player_sprite.top

        # Add the bullet to the appropriate lists
        self.bullet_list.append(bullet)

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on bullet sprites
        self.bullet_list.update()

        # Loop through each bullet
        for bullet in self.bullet_list:

            # Check this bullet to see if it hit a coin
            hit_list = arcade.check_for_collision_with_list(bullet, self.coin_list)

            # If it did, get rid of the bullet
            if len(hit_list) > 0:
                bullet.remove_from_sprite_lists()

            # For every coin we hit, add to the score and remove the coin
            for coin in hit_list:
                coin.remove_from_sprite_lists()
                self.score += 1

                # Hit Sound
                arcade.play_sound(self.hit_sound)

            # If the bullet flies off-screen, remove it.
            if bullet.bottom > WINDOW_HEIGHT:
                bullet.remove_from_sprite_lists()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()
